
// Application CompanyInfo - Contains info about a company
//
//  Public Interfaces:
//
//    CompanyInfo() - Constructor


//    set(value, attribute) - set a value for an attribute
//    set(value, attribute, year) - set a value for an attribute for a
//		specified year
//    set(value, attribute, year, per) - set a value for an attribute for
//		a specified year and period
//    set(array_of_items) - set a collection of values
//		Each element of the array is itself an array of
//		    2 items - value & attribute
//		    3 items - value, attribute & year
//		or  4 items - value, attribute, year & period

//    get(attribute) - get a value for an attribute
//    get(attribute, year) - get a value for an attribute for a
//		specified year
//    get(attribute, year, per) - get a value for an attribute for
//		a specified year and period

//    dump() - for debugging; dump all settings out

// ----------------------------------------------------------------

//  All values & attributes are arbitrary content strings; e.g. "34.5",
//    "Outstanding Shares", etc.
//  Years are strings of the year number, e.g. "1998" etc.
//  Periods are strings made up a letter and a number (case is not relevant);
//    e.g. "Q1", "m3" - 1st quarter, third month, etc.
//    With the allowable indicator letters as:
//	Q - Quarter (1-4)
//	H - Half (1-2)
//	M - Month (1-12)
//	W - Week (1-52)
//	D - Day (1-365)

package sptossg ;

import java.util.*;
import java.io.*;

public class CompanyInfo implements CompanyStorage {



    // ****************************************************************
    // Data structures initialization & processing

    // Constructor
    public CompanyInfo () {
	baseperiods = new Hashtable();
	minperiods = new Hashtable();
	maxperiods = new Hashtable();
	datasets = new Hashtable();
    }

    // Storage for all datasets, collections of data characterized by
    // dates; e.g. annual sales, quarterly dividends, etc.
    //
    // the key to this hashtable is a string representing the attribute
    // with the value being the DataSet for that attribute
    Hashtable datasets;

    // The key to be used for static value datasets
    static final Period always = new Period();

    // getDataSet - retrieve the dataset for a specified attribute.  If a
    // dataset hasn't yet been created for the attribute, it will create it
    // if build is true.  If there is a defined DynamicDataSet (a DataSet
    // with a formula to calculate the value) then it will add a copy to
    // the company.
    DataSet getDataSet(String str, boolean build) {
	Object set;

	// First check for any instantiated dataset
	set = datasets.get(str);

	// if none, then build a regular DataSet
	if (set == null && build == true) {
	    DataSet ds = new DataSet(str);
	    datasets.put(str, ds);
	    set = ds;
	}

	if (set instanceof DataSet) {
	    return (DataSet) set;
	} else {
	    // throw new incompatibleFormatException();
	    return  null;
	}
    }


    Hashtable baseperiods;
    Hashtable minperiods;
    Hashtable maxperiods;

    // set the base period for attributes of a specific frequency to handle
    // relative periods properly
    public void setBasePeriod(int year, int yearpart, int frequency) {
	Period baseperiod = new Period(year, yearpart, frequency);

	Integer freqkey = new Integer(frequency);

	baseperiods.put(freqkey, baseperiod);

	Enumeration e = datasets.keys();
	while (e.hasMoreElements()) {
	    String key = (String) e.nextElement();
	    DataSet ds = (DataSet) datasets.get(key);
	    if (ds.frequency() == frequency) {
		ds.setBasePeriod(this, baseperiod);
	    }
	}
    }

    public void clearBasePeriods() {
	baseperiods = new Hashtable();
    }


    public Period updateBasePeriod(Period p) {
	int frequency = p.frequency();
	Integer freqkey = new Integer(frequency);

	Period maxper = (Period) baseperiods.get(freqkey);

	if (p.greater(maxper)) {
	    baseperiods.put(freqkey, p);
	    return p;
	}
	return maxper;
    }

    public Period getBasePeriod(int frequency) {
	Integer freqkey = new Integer(frequency);
	Period baseperiod = (Period) baseperiods.get(freqkey);
	return baseperiod;
    }

    public Period getMaxPeriod(int frequency) {
	Integer freqkey = new Integer(frequency);
	Period maxperiod = (Period) maxperiods.get(freqkey);
	return maxperiod;
    }

    public void setMaxPeriod(Period p) {
	int frequency = p.frequency();
	Integer freqkey = new Integer(frequency);

	Period maxper = (Period) maxperiods.get(freqkey);

	if (p.greater(maxper)) {
	    maxperiods.put(freqkey, p);
	}
    }

    public Period getMinPeriod(int frequency) {
	Integer freqkey = new Integer(frequency);
	Period minperiod = (Period) minperiods.get(freqkey);
	return minperiod;
    }

    public void setMinPeriod(Period p) {
	int frequency = p.frequency();
	Integer freqkey = new Integer(frequency);

	Period minper = (Period) minperiods.get(freqkey);

	if (p.less(minper)) {
	    minperiods.put(freqkey, p);
	}
    }


    // ****************************************************************
    // Storage methods

    // value set using period - actual workhorse set method
    public String set(String value, String attr, Period per) {
	boolean build = true;
	if (value.equals("") || value == null) { build = false; }
	DataSet ds = getDataSet(attr, build);
	if (ds == null) { return null; }
	ds.store(this, value, per);
	return value;
    }

    // value set for static values - period is 'always'
    public String set(String value, String attr) {
	return set(value, attr, always);
    }

    // value set for annual data - year part is "Y1" (entire year)
    public String set(String value, String attr, String year) {
	return set(value, attr, year, "Y1");
    }

    // value set for periodic data (quarterly, etc.), using string parameters
    public String set(String value, String attr,
		      String year, String yearpart) {
	return set(value, attr, new Period(year, yearpart));
    }

    // value set for periodic data, using numeric parameters
    public String set(String value, String attr,
		      int year, int yearpart, int frequency) {
	return set(value, attr, new Period(year, yearpart, frequency));
    }

    // value set for array of data (attr, year, period), all items as strings
    public String set(String[][] items) {

	for (int i = 0; i < items.length; i++) {
	    String itm[] = items[i];

	    String val = itm[0];
	    String label = itm[1];

	    if (itm.length == 2) {
		set(val, label);
	    } else if (itm.length == 3) {
		String yr = itm[2];
		set(val, label, yr);
	    } else {
		String yr = itm[2];
		String per = itm[3];
		set(val, label, yr, per);
	    }
	}
	return "";
    }

    // ****************************************************************
    // Retrieval methods

    // value retrieval using period - actual workhorse retrieval method
    //
    // This method gets a requested value in the following priority order
    //   1. A value set for this explicit time period
    //   2. A formula that can calculate the value for this particular
    //      time period
    //   3. A value set for this attribute for any time period
    //   4. A formula that can calculate this value for any time period.
    public String get(String attr, Period per) {
	DataSet ds = getDataSet(attr, false);
	String retval = null;

	// first check for an explicit set value for this period
	if (ds != null) { retval = ds.retrieve(this, per); }
	if (retval != null) { return retval; }

	// if no explicit value for date, check for a dynamic dataset formula
	// that calculates a value for the date
	retval = DynamicDataSet.retrieve(this, attr, per);
	if (retval != null) { return retval; }

	// otherwise let's also check for static value (within this company)
	// for retrieval
	if (per != always && ds != null) { retval = ds.retrieve(this, always); }
	if (retval != null) { return retval; }

	// if no explicit value for any date, check for a dynamic dataset formula
	// that calculates a value for any date

        // **** This may be irrelevant in that the previous DynamicDataSet
        // retrieval (for a specific date) would probably have returned
        // formula calculations for any date.
	if (per != always) {
	    retval = DynamicDataSet.retrieve(this, attr, always);
	}
	if (retval != null) { return retval; }

	// Check for global (across all companies) values
	if (globalcompany != null) {
	    retval = globalcompany.get(attr, per);
	}
	if (retval != null) { return retval; }

	return null;
    }

    // Storage for any global settings of attributes
    static CompanyInfo globalcompany = null;

    // value retrieval for static data - period is 'always'
    public String get(String attr) {
	return get(attr, always);
    }

    // value retrieval for annual data - year part is "Y1" (entire year)
    public String get(String attr, String year) {
	return get(attr, year, "Y1");
    }

    // value retrieval for periodic data, string parameters
    public String get(String attr, String year, String yearpart) {
	return get(attr, new Period(year, yearpart));
    }

    // value retrieval for periodic data, numeric parameters
    public String get(String attr, int year, int yearpart, int frequency) {
	return get(attr, new Period(year, yearpart, frequency));
    }

    // ****************************************************************
    // Routines for testing

    public void dumpAttrs() {
	Enumeration e;

	int i = 0;

	e = datasets.keys();
	System.out.println("Company Attributes: ");
	while (e.hasMoreElements()) {
	    String key = (String) e.nextElement();
	    DataSet ds = (DataSet) datasets.get(key);
	    int freq = ds.frequency();
	    String freqstr;
	    if (freq == 0) {
		freqstr = "unknown";
	    } else {

		freqstr = Period.periodicString(freq);
	    }
	    System.out.println(i++ + ": (" + freqstr + ") " + key);
	}
    }

    // test_sets - a method demonstrating how to use the various public
    // interface methods (set, get) to store and retrieve values
    void test_sets() {
	// Example SETs, done individually
	set("SSG Creators, Inc.", "Name");
	set("58.6", "Annual Sales", "2001");
	set("10.2", "Quarterly Income", "1999", "Q3");

	// Example SETs, via array
	String tostore[][] = {
	    {"BIBH", "Ticker"},
	    {"47.1", "Annual Sales", "1999"},
	    {"53.4", "Annual Sales", "2000"},
	    {"92.0", "Annual Sales", "2002"},

	    {"11.3", "Annual Profit", "1999"},
	    {"19.4", "Annual Profit", "2000"},
	    {"24.6", "Annual Profit", "2001"},
	    {"44.4", "Annual Profit", "2002"},

	    {"8.9", "Quarterly Income", "1999", "Q1"},
	    {"9.4", "Quarterly Income", "1999", "Q2"},
	    {"11.4", "Quarterly Income", "1999", "Q4"},

	    {"18.9", "Quarterly Income", "2000", "Q1"},
	    {"19.4", "Quarterly Income", "2000", "Q2"},
	    {"20.2", "Quarterly Income", "2000", "Q3"},
	    {"21.4", "Quarterly Income", "2000", "Q4"},
	};
	set(tostore);


	// Example GETs
	String name = get("Name");
	System.out.println("Name = '" + name + "'.");

	String sales2000 = get("Annual Sales", "2000");
	System.out.println("Sales(2000) = '" + sales2000 + "'.");

	String qi2000q1 = get("Quarterly Income", "2000", "Q1");
	System.out.println("Quarterly Income(2000,Q1) = '" + qi2000q1 + "'.");

	// Some more GETs, by array; some aren't set
	String tofetch[][] = {
	    {"Annual Sales", "2001"},
	    {"Quarterly Income", "1999", "Q2"},
	    {"Quarterly Income", "1989", "Q1"},
	    {"Annual Profit", "2002"},
	    {"Annual Profit", "2004"}};

	for (int i = 0; i < tofetch.length; i++) {
	    String itm[] = tofetch[i];

	    String label = itm[0];
	    String yr = itm[1];

	    if (itm.length == 2) {
		String val = get(label, yr);
		System.out.println(label + " for " + yr + " = " + val);
	    } else {
		String per = itm[2];

		String val = get(label, yr, per);
		System.out.println(label + " for " + yr + "/" + per + " = " + val);
	    }
	}

	System.out.println();

	dump();
    }

    public Vector getKeys(String attr) {
	DataSet ds = (DataSet) datasets.get(attr);
	if (ds != null) { return ds.getKeys(this); }
	return DynamicDataSet.getKeys(this, attr);
    }

    public void dump(String attr) {
	DataSet ds = (DataSet) datasets.get(attr);
	if (ds != null) {
	    ds.dump(this, attr);
	} else {
	    DynamicDataSet.dump(this, attr);
	}
    }

    // dump - print out all attributes and their values
    public void dump() {
	Enumeration e;

	System.out.println("\nStatic Values: ");
	e = datasets.keys();
	while (e.hasMoreElements()) {
	    String attr = (String) e.nextElement();
	    DataSet ds = (DataSet) datasets.get(attr);
	    if (! ds.isPeriodic()) {
		String val = ds.retrieve(this, always);

		System.out.println("    " + attr + ": " + val);
	    }
	}

	System.out.println("Periodic Values: ");
	e = datasets.keys();
	while (e.hasMoreElements()) {
	    String attr = (String) e.nextElement();
	    DataSet ds = (DataSet) datasets.get(attr);
	    if (ds.isPeriodic()) {
		ds.dump(this, attr);
	    }
	}

	DynamicDataSet.dump(this);
    }

    public void dprint() {
	System.out.println("this = '" + this + "'.");
	Debug.dumpHashtable(baseperiods, "Base Periods:");
	Debug.dumpHashtable(minperiods, "Min Periods:");
	Debug.dumpHashtable(maxperiods, "Max Periods:");
	Debug.dumpHashtable(datasets, "Datasets:");
    }

    public void dprint (String attr) {
	DataSet ds = (DataSet) datasets.get(attr);
	if (ds != null) {
	    ds.dprint(attr);
	} else {
	    DynamicDataSet.dprint(attr);
	}
    }

    // ****************************************************************
    // Data file format descriptions


    // fileFormatItem - an inner class that defines an item that goes into
    // a data output file.  There are three types of items;
    //   CONSTANTS - A string written into the file as is;
    //      e.g. "Version 7", "," etc.
    //   a STATIC - a non-dated data item
    //      e.g. "Company Name", "Company Ticker"
    //   a PERIODIC - a dated item, relative to the most recent date stored;
    //      e.g. "Annual Sales" for 3 years ago, "Quarerly EPS" for 8 quarters ago;
    //
    // each item contains a count, interpreted as follows:
    //  for CONSTANTS or STATICS - value is inserted N times
    //  for PERIODICS - the OFFSET specifies the start period; the INCREMENT specifies
    //   how to get the next one;

    // CALCULATED may be used later for attributes that are calculated
    // from other attributes
    // static public final int CALCULATED = 0;

    // CONSTANT values are specified explicitly, not retrieved based on company
    static public final int CONSTANT = 0;


    // STATIC (or SINGLE) values have one value per company
    static public final int STATIC = 1;
    static public final int SINGLE = 1;

    // PERIODIC data have values associated with date periods (years,
    // quarters, etc.)
    static public final int PERIODIC = 2;
    static public final int ANNUAL = 3;
    static public final int QUARTERLY = 4;
    static public final int MONTHLY = 5;

    static class fileFormatItem {

	String label;
	int itmType;
	int count;

	int frequency;

	int offset;
	int increment;

	public fileFormatItem(String label) {
	    this(label, SINGLE);
	};

	public fileFormatItem(String label, int itmType) {
	    this(label, itmType, 1, 0, 0);
	};

        public fileFormatItem(String label, int itmType, int count, int
			      offset, int increment) {
	    this.label = label;
	    if (itmType == ANNUAL) {
		this.frequency = 1;
		this.itmType = PERIODIC;
	    } else if (itmType == QUARTERLY) {
		this.frequency = 4;
		this.itmType = PERIODIC;
	    } else if (itmType == MONTHLY) {
		this.frequency = 12;
		this.itmType = PERIODIC;
	    } else {
		this.itmType = itmType;
	    }
	    this.count = count;
	    this.offset = offset;
	    this.increment = increment;
	}
    }

    // This array describes the format of an SSG file.  made up of 537
    // entries (if there are only 400 in the file, then the first 400 items
    // define the older SSG format).

    static fileFormatItem format_SSG[] = {

	new fileFormatItem("High Price", ANNUAL, 10, -9, 1),
	new fileFormatItem("Low Price", ANNUAL, 10, -9, 1),
	new fileFormatItem("Sales", ANNUAL, 10, -9, 1),
	new fileFormatItem("Tax Rate", ANNUAL, 10, -9, 1),
	new fileFormatItem("PreTax Inc", ANNUAL, 10, -9, 1),
	new fileFormatItem("Net Income", ANNUAL, 10, -9, 1),
	new fileFormatItem("Book Value", ANNUAL, 10, -9, 1),
	new fileFormatItem("Earnings per Share", ANNUAL, 10, -9, 1),
	new fileFormatItem("Dividend per Share", ANNUAL, 10, -9, 1),

	// 90 items, annual data

	new fileFormatItem("Current Price"),
	new fileFormatItem("Current P/E"),
	new fileFormatItem("Current Dividend"),
	new fileFormatItem("Last Year of data"),
	new fileFormatItem("Number of Data Years Used"),
	new fileFormatItem("52 Week High Price"),
	new fileFormatItem("52 Week Low Price"),
	new fileFormatItem("Number of Common Shares Outstanding"),
	new fileFormatItem("Number of Authorized Common Shares"),
	new fileFormatItem("Number of Preferred Shares Outstanding"),

	// 100 items

	new fileFormatItem("Number of Authorized Preferred Shares"),
	new fileFormatItem("Total Debt"),
	new fileFormatItem("Potential Dilution"),
	// ????????????????
	// The following two items appear to be duplicates of
	// parts of the items at rows 202 & 210.
	new fileFormatItem("Quarterly Sales", QUARTERLY, 2, 0, -4),
	new fileFormatItem("Quarterly E/S", QUARTERLY, 2, 0, -4),
	new fileFormatItem("Last CY of data +5"),
	new fileFormatItem("Outside E/S Estimate"),
	new fileFormatItem("Choice Low Price (5-Year Average Low)"),

	// 110 items

	new fileFormatItem("Choice High PE"),
	new fileFormatItem("Choice Low PE"),
	new fileFormatItem("High Yield"),
	new fileFormatItem("Add/Delete flag", SINGLE, 10, 0, 0),
	new fileFormatItem("Low E/S Estimate"),
	new fileFormatItem("Quarterly E/S Growth"),
	new fileFormatItem("Projected E/S Growth"),
	new fileFormatItem("Quarterly Sales Growth"),
	new fileFormatItem("Projected Sales Growth"),
	// ????????????????
	// Current Yield - appears to be duplicated at line 354
	new fileFormatItem("Current Yield"),
	new fileFormatItem("Average Yield"),

	// 130 items

	new fileFormatItem("Historic E/S Growth"),
	new fileFormatItem("Average PE"),
	new fileFormatItem("U:D ratio"),
	new fileFormatItem("5-Year Sales Estimate"),
	new fileFormatItem("Average Pre-Tax Profit"),
	new fileFormatItem("Income Tax Rate"),
	new fileFormatItem("Preferred Dividends"),
	new fileFormatItem("Projected Liabilities"),
	new fileFormatItem("Projected Outstanding Shares"),
	new fileFormatItem("Prefferred Procedure E/S"),

	// 140 items

	new fileFormatItem("Visual E/S Estimate"),
	new fileFormatItem("Last CY Trend E/S"),
	new fileFormatItem("Historic Sales Growth"),
	new fileFormatItem("Average ROE"),
	new fileFormatItem("Low Price Estimate"),
	new fileFormatItem("Average Price Estimate"),
	new fileFormatItem("Sell Level Price"),
	new fileFormatItem("Projected Dividend Payout %"),
	new fileFormatItem("Previous E/S"),
	new fileFormatItem("New E/S Estimate"),

	// 150 items

	new fileFormatItem("Projected High Price"),
	new fileFormatItem("Recent Cash on Hand", ANNUAL), // Was "Current Cash"
	new fileFormatItem("Accounts Receivables", ANNUAL),
	new fileFormatItem("Inventories", ANNUAL),
	new fileFormatItem("Assets", ANNUAL),
	new fileFormatItem("Short Term Debt (Sum Last 4 Qtrs)", ANNUAL),
	new fileFormatItem("Liabilities", ANNUAL),
	new fileFormatItem("Cash", ANNUAL, 1, -1, 0),
	new fileFormatItem("Accounts Receivables", ANNUAL, 1, -1, 0),
	new fileFormatItem("Inventories", ANNUAL, 1, -1, 0),

	// 160 items

	new fileFormatItem("Assets", ANNUAL, 1, -1, 0),
	new fileFormatItem("Short Term Debt", ANNUAL, 1, -1, 0),
	new fileFormatItem("liabilities", ANNUAL, 1, -1, 0),
	new fileFormatItem("Total Capitalization"),
	new fileFormatItem("Long Term Debt", ANNUAL, 10, -9, 1),
	new fileFormatItem("Number shares outstanding", ANNUAL, 10, -9, 1),
	new fileFormatItem("Cash flow/share", ANNUAL, 10, -9, 1),
	new fileFormatItem("Total Assets"),
	new fileFormatItem("Interest Coverage %"),
	new fileFormatItem("Total Return"),
	new fileFormatItem("Projected E/S Growth"),
	new fileFormatItem("Severe Low Price"),
	new fileFormatItem("R Squared"),

	// 200 Items

	new fileFormatItem("Reserved"),
	new fileFormatItem("Quarterly Sales", QUARTERLY, 8, -7, 1),
	new fileFormatItem("Quarterly E/S", QUARTERLY, 8, -7, 1),
	new fileFormatItem("Quarterly Dividends", QUARTERLY, 8, -7, 1),
	new fileFormatItem("Inside Ownership"),
	new fileFormatItem("Institutional Holdings"),
	new fileFormatItem("Reserved"),
	new fileFormatItem("ACE 5 Year Estimate E/S Growth Rate"),
	new fileFormatItem("Predictability"),
	new fileFormatItem("Long-Term Debt"),
	new fileFormatItem("Fiscal Year End Month"),
	new fileFormatItem("Reserved", SINGLE, 3, 0, 0),
	new fileFormatItem("Name"),
	new fileFormatItem("Symbol"),
	new fileFormatItem("Exchange"),
	new fileFormatItem("Data Source"),
	new fileFormatItem("Date of Data Creation"),

	// 240 items

	new fileFormatItem("Reserved"),	// was "V-L page"
	new fileFormatItem("Code"),
	new fileFormatItem("Reserved"),
	new fileFormatItem("Industry"),
	new fileFormatItem("Split Ratio During Quarter"),
	new fileFormatItem("Reserved", SINGLE, 9, 0, 0),
	new fileFormatItem("Latest Quarter Ending Date"),
	new fileFormatItem("Reserved", SINGLE, 14, 0, 0),
	new fileFormatItem("Month of Latest Quarterly Data"),

	// 270 items

	new fileFormatItem("Year of Latest Quarterly Data"),
	new fileFormatItem("Reserved", SINGLE, 4, 0, 0),
	new fileFormatItem("Software Version"),
	new fileFormatItem("Reserved", SINGLE, 2, 0, 0),
	new fileFormatItem("SIC Code"),
	new fileFormatItem("S & P Stock Rank"),

	// 280 items

	new fileFormatItem("Financial Strength VL"),
	new fileFormatItem("Safety VL"),
	new fileFormatItem("Reserved", SINGLE, 3, 0, 0),
	new fileFormatItem("Beta"),
	new fileFormatItem("Reserved", SINGLE, 28, 0, 0),
	new fileFormatItem("Quarterly Pre-Tax Income", QUARTERLY, 8, -7, 1),
	new fileFormatItem("Quarterly Net Income", QUARTERLY, 8, -7, 1),

	// 330 items

	new fileFormatItem("Last Monthly Closing Price Date"),
	new fileFormatItem("Reserved", SINGLE, 15, 0, 0),
	new fileFormatItem("E/S in 5 Yrs. Based on Ace Proj Growth Rate"),
	new fileFormatItem("Reserved"),		// was "Historical 5 Yr E/S Growth Rate"
	new fileFormatItem("Reserved"),		// was "Current PE/Ace 5 Yr Growth rate"
	new fileFormatItem("S&P Rating"),

	// 350 items

	new fileFormatItem("Reserved"),		// was "Avg 5 Yr PE"
	new fileFormatItem("Reserved"),		// was "Proj 5 Yr High Price"
	new fileFormatItem("Reserved"),		// was "Proj Compounded Appreciation of Price"
	new fileFormatItem("Current Yield"),
	new fileFormatItem("Reserved"),		// was "Proj Total Return"
	new fileFormatItem("Historical 5 Yr. Sales Growth Rate"),
	new fileFormatItem("Historical 5 Yr E/S Growth R-Squared"),
	new fileFormatItem("Industry Group"),
	new fileFormatItem("Reserved"),		// was "% Off 52 Week High Price"
	new fileFormatItem("Reserved"),		// was "% Above 52 Week Low Price"

	// 360 items

	new fileFormatItem("Reserved"),		// was "Relative Value"
	new fileFormatItem("Reserved"),		// was "PE/Avg ROE"
	new fileFormatItem("Reserved"),		// was "Debt as % of Capital"
	new fileFormatItem("Next Year E/S estimate"),
	new fileFormatItem("Historical 5 Yr. Sales Growth R-Squared"),
	new fileFormatItem("Reserved", SINGLE, 35, 0, 0),

	// 400 items

	// Next 137 Items are for the long SSG file

	new fileFormatItem("Sector"),
	new fileFormatItem("Quarterly Sales", QUARTERLY, 12, -19, 1),  // Verify translation!
	new fileFormatItem("Quarterly E/S", QUARTERLY, 12, -19, 1),  // Verify translation!

	// 425 Items

	new fileFormatItem("Quarterly Pre-Tax Income", QUARTERLY, 12, -19, 1),
	new fileFormatItem("Common Shares Outstanding", QUARTERLY, 20, -19, 1),
	new fileFormatItem("Tax Rate Qtrly", QUARTERLY, 20, -19, 1),

	// 477 Items

	new fileFormatItem("Closing Price", MONTHLY, 60, -59, 1)

	// 537 Items
    };

    // ****************************************************************
    // Data file I/O routines

    // Data file writing

    void updateBasePeriods(fileFormatItem[] items) {
	clearBasePeriods();
	for (int i = 0; i < items.length; i++) {
	    Period p = getLatestPeriod(items[i]);
	    if (p != null) {
		updateBasePeriod(p);
	    }
	}
    }

    Period getLatestPeriod(fileFormatItem item) {

	// Only handle periodic items
	if (item.itmType != PERIODIC) { return null; }

	// Start at the max and count down till a value is found
	Period max = getMaxPeriod(item.frequency);
	Period min = getMinPeriod(item.frequency);
	if (max == null || min == null) { return null; }
	Period cur = max;
	String val = null;
	while (val == null && (! min.greater(cur))) {
	    val = get(item.label, cur);
	    if (val == null) { cur = cur.adjust(-1); }
	}
	if (val == null) { return null; }
	return cur;
    }

    // writeSSGfile - write out all data in SSG format to the specified file
    public void writeSSGfile(String fname) {
	// Find the appropriate base periods for the items to be written
	// to the output file
	updateBasePeriods(format_SSG);
	// Build relative dates for annual data
	store_relative_date(1, "[Last Year of data]", "year()");
	// Build relative dates for quarterly data
	store_relative_date(4, "[Year of Latest Quarterly Data]", "year()");
	store_relative_date(4, "[Month of Latest Quarterly Data]",
			    "1 + (3*period() + [Fiscal Year End Month]) % 12");
	// Build relative dates for monthly data
	// store_relative_date(12, 
	writeData(fname, format_SSG, "\n", "");
    };

    // !!!!!!!!!!!!!!!!
    // Note: there's a conceptual problem here.  The data is kept in
    // periods in a numeric form, but may be stored in the files in some
    // form of string form, though Formula values don't do non-numerics.

    void store_relative_date (int frequency, String attr, String formula) {
	Period base = getBasePeriod(frequency);
	if (base == null) {
	    System.out.println("No data stored for period frequency " + frequency);
	    return;
	}
	try {
	    double val = Formula.eval(formula, this, base);
	    if (Double.isNaN(val)) { return; }
	    set("" + val, attr);
	} catch (Exception exc) {
	    System.out.println("Error evaluating relative date formula for " +
			       attr + " = '" + formula + "'");
	    System.out.println("Error was " + exc);
	}
    }
 
    // writeData - actually write the data to a file.  Parameters are:
    // fname - the file name to write out
    // items - an array of fileFormatItem that describe what needs to be written out
    // sep - the separator to be written between items
    // ifnull - what should be written out if a value is NULL
    void writeData(String fname, fileFormatItem[] items, String sep, String ifnull) {

	PrintWriter file;
	try {
	    file = new PrintWriter(new FileOutputStream(fname));
	} catch (Exception exc) {
	    System.out.println("Error in opening file " + fname);
	    return;
	}
	for (int i = 0; i < items.length; i++) {
	    writeItem(file, items[i], sep, ifnull);
	}
	file.close();
    }

    // write an individual data attribute to the output file
    void writeItem (PrintWriter file, fileFormatItem item,
		    String sep, String ifnull) {
	String key = item.label;
	String val;

	switch (item.itmType) {
	case CONSTANT:
	    val = key;
	    for (int i = 0; i < item.count; i++) {
		file.print(val + sep);
	    }
	    break;
	case STATIC:
	    val = get(key, always);
	    if (val == null) { val = ifnull; }
	    for (int i = 0; i < item.count; i++) {
		file.print(val + sep);
	    }
	    break;
	case PERIODIC:
	    Period per = getBasePeriod(item.frequency);
	    if (per == null) {
		// if no baseperiod for this frequency level, then values
		// should still be stored in relative offset format
		per = new Period(0, item.offset, item.frequency);
	    } else {
		per = per.adjust(item.offset);
	    }

	    for (int i = 0; i < item.count; i++) {
		val = get(key, per);
		if (val == null) { val = ifnull; }
		file.print(val + sep);
		per = per.adjust(item.increment);
	    }
	    break;
	default:
	    System.out.println("Unknown format type " + item.itmType +
			       " for file item " + key);
	}}

    // Data file loading

    // loadSSGfile - read in all data in SSG format from the specified file
    public void loadSSGfile(String fname) {
	loadData(fname, format_SSG, "\n", "");
	// convert relative annual data to absolute
	apply_relative_date(1, "[Last Year of data]", "1");
	// convert relative quarterly data to absolute
	String qtly_per_fn =
	    "1 + (4 + floor([Month of Latest Quarterly Data] / 3)" +
	    "       - floor([Fiscal Year End Month] / 3)" +
	    "       - 1) % 4";
	apply_relative_date(4, "[Year of Latest Quarterly Data]", qtly_per_fn);
	// convert relative monthly data to absolute
	// ????????????????
	System.out.println("!!!! Don't know how to retrieve the " +
			   "monthly base date.");
	// apply_relative_date(12, "monthly date yr", "monthly date month");
    };

    // apply_relative_date converts all relative dates into absolute dates
    // by evaluating the formulas to calculate out the absolute year and
    // period references for a particular frequency, and setting the base
    // date for that frequency to be that referent.
    void apply_relative_date(int frequency, String absyearfn, String absperfn) {
	int year;
	int period;

	try {
	    year = (int) Formula.eval(absyearfn, this);
	} catch (Exception exc) {
	    System.out.println("Error: Can't evaluate year function " +
			       "for frequency = " + frequency);
	    System.out.println("Function = '" + absyearfn + "'.");
	    return;
	}

	try {
	    period = (int) Formula.eval(absperfn, this);
	} catch (Exception exc) {
	    System.out.println("Error: Can't evaluate period function " +
			       "for frequency = " + frequency);
	    System.out.println("Function = '" + absperfn + "'.");
	    return;
	}

	try {
	    setBasePeriod(year, period, frequency);
	} catch (Exception exc) {
	    System.out.println("Error: Can't set dates for frequency = " +
			       frequency);
	}
    }

    // load input data from a file
    void loadData(String fname, fileFormatItem[] items,
		  String sep, String ifnull) {

	BufferedReader file;
	try {
	    FileInputStream fis = new FileInputStream(fname);
	    file = new BufferedReader(new InputStreamReader(fis));
	} catch (Exception exc) {
	    System.out.println("Error in opening file " + fname);
	    return;
	}

	for (int i = 0; i < items.length; i++) {
	    try {
		loadItem(file, items[i], sep, ifnull);
	    } catch (Exception exc) {
		System.out.println("Error: " + exc);
		System.out.println("File truncated at " + i + " data items. ");
		try {
		    file.close();
		} catch (Exception exc2) {}
		return;
	    }
	}
	try {
	    file.close();
	} catch (Exception exc) {}
    }

    void loadItem (BufferedReader file, fileFormatItem item,
		   String sep, String ifnull)
	throws IOException {
	String key = item.label;
	String val;
	switch (item.itmType) {
	case CONSTANT:
	    for (int i = 0; i < item.count; i++) {
		val = file.readLine();
	    }
	    break;
	case STATIC:
	    for (int i = 0; i < item.count; i++) {
		val = file.readLine();
		set(val, key);
	    }
	    break;
	case PERIODIC:
	    int period = item.offset;
	    int incr = item.increment;
	    for (int i = 0; i < item.count; i++) {
		val = file.readLine();
		set(val, key, 0, period, item.frequency);
		period += incr;
	    }
	    break;
	default:
	    System.out.println("Unknown format type " + item.itmType +
			       " for file item " + key);
	}}
}
