
// Class DynamicDataSet

// This class defines datasets whose values can be calculated.
//  e.g.
//     Outstanding Shares <=> "Net Profits" divided by "EPS"

// This class extends the DataSet class as follows:

// The constructor takes a Formula for use in calculating values

// The semantics of the storage methods are the same as in DataSet and store
// values explicitly.

// The semantics of retrieval is that if an explicit value is stored, then
// that will be used.  If no explicit value is available, then the formula is
// used to calculate the result value.

// Usage:
//    The constructors of DynamicDataSet are private methods and not
//    accessible to other classes.  Instead, there is a public class method
//    'define(name, formula-definition)' to be used.
//    Example Usage:
//        DynamicDataSet.define("PTP Margin", "100 * [PreTax Inc] / Sales");

// Also, DynamicDataSets are not defined on a per-company basis, but
// instead are defined generally.  Retrievals of a DDS for a company will
// instantiate that DDS within that company the first time a retrieval is
// attempted.

package sptossg ;

import java.util.*;

public class DynamicDataSet {

    Formula formula;
    String name;

    // Constructor for a DynamicDataSet - Note that this is private;
    // all definitions of DynamicDataSets should go through the public
    // static method 'define'.
    private DynamicDataSet (String name, Formula formula) {
	this.name = name;
	this.formula = formula;
	if (curGroup != null) {
	    curGroup.store(name, this);
	} else {
	    System.out.println("!!!!  Error in DDS groups.  No current group! ");
	}
    }

    static public DynamicDataSet define(String name, String def) {

	char c1 = name.charAt(0);

	if ((c1 >= 'a' && c1 <= 'z') || (c1 >= 'A' && c1 <= 'Z')) {
	    Formula formula = Formula.parse(def);
	    if (formula == null) { return null; }
	    return new DynamicDataSet(name, formula);
	} else {
	    System.out.println("Invalid dynamic attribute name '" + name + "'. ");
	    return null;
	}
    }

    // static method to find a named DynamicDataSet
    static DynamicDataSet find(String name) {
	DynamicDataSet set = findDDS(name);
	return set;
    }

    static public void dump(CompanyStorage company) {
	System.out.println("Calculated Values: ");

	for (int i = 0; i < groups.size(); i++) {
	    ddsGroup curgrp = (ddsGroup) groups.elementAt(i);
	    dump(company, curgrp);
	}
    }

    static public void dump (CompanyStorage company, ddsGroup curgrp) {
	System.out.println("  Group " + curgrp.name +
			   ": Priority: " +  curgrp.priority +
			   ", " + (curgrp.active ? "Active" : "Inactive"));
	Enumeration e = curgrp.defs.keys();
	while (e.hasMoreElements()) {
	    String name = (String) e.nextElement();
	    DynamicDataSet dds = (DynamicDataSet) curgrp.defs.get(name);
	    if (dds != null) {
		dds.dumpVals(company, name);
	    } else {
		Debug.debug("No Dynamic dataset found for " + name);
	    }
	}
    }

    static public void dump (CompanyStorage company, String name) {
	DynamicDataSet dds = findDDS(name);
	dds.dumpVals(company, name);
    }

    void dumpVals (CompanyStorage company, String name) {

	String indent = "    ";
	System.out.println(indent + name + ": Dynamic DataSet");
	indent += indent;

	Vector keys;

	// Check if it's a single variabled Dynamic DataSet
	Period p = new Period();
	String val = retrieve(company, p);
	if (val != null) {
	    keys = new Vector();
	    keys.addElement(p);
	} else {
	    keys = formula.getKeys(company);
	    if (keys == null) {
		System.out.println(indent + "No keys found. ");
		return;
	    }
	}

	Enumeration e = keys.elements();
	while (e.hasMoreElements()) {
	    p = (Period) e.nextElement();

	    val = retrieve(company, p);

	    // if we have a value for this key
	    if (val != null) {
		double nval = (new Double(val)).doubleValue();
		// round to two digits
		int nval2 = (int) (nval * 100);
		val = "" + (nval2 / 100.0);

		System.out.println(indent + p.toString() + ": " + val);
	    }
	}
    }

    static public void dprint (String name) {
	DynamicDataSet dds = findDDS(name);
	if (dds != null) {
	    dds.dprint();
	} else {
	    System.out.println("No DynamicDataSet found for name = '" +
			       name + "'. ");
	}
    }

    public void dprint () {
	System.out.println("this = '" + super.toString() + "'.");
	System.out.println("name = '" + name + "'.");
	System.out.println("formula = '" + formula + "'.");

    }

    public String toString() {
	return ("def([" + name + "]) = \"" + formula + "\"");
    }

    // External retrieve routine -
    //  Static method called with dynamic attribute name and period
    //  finds the right DynamicDataSet and then evaluates that
    //  DynamicDataSet for that period.
    static public String retrieve(CompanyStorage company, String name, Period p) {
	DynamicDataSet dds = findDDS(name);
	if (dds == null) { return null; }
	return dds.retrieve(company, p);
    }

    // doing_evaluation is used for automatic cycle detection during
    // evaluation.  For example, if A = B / 10 and B = A * 10, an
    // evaluation of either A or B will cause a undesired cycle.
    boolean doing_evaluation = false;

    //  use formula to get value
    String retrieve(CompanyStorage company, Period p) {
	double val;
	String result;

	// if we're already in an evaluation of this dynamic attribute,
	// then terminate the cycle with no value.  Note that this
	// cycle detection should allow us to define DDS formulas based
	// on one another and do retrievals as long as one of the
	// attributes has an explicit value stored.
	if (doing_evaluation) {
	    System.out.println("Warning: Evaluation cycle detected on " + name);
	    val = Double.NaN;
	} else {
	    try {
		doing_evaluation = true;
		val = formula.evaluate(company, p);
	    } finally {
		doing_evaluation = false;
	    }
	}
	if (Double.isNaN(val)) { result = null; }
	else { result = ("" + val); }
	return result;
    }

    // getKeys - Return all keys (the set of periods that might be valid
    // for this dynamic attribute), in sorted order
    Vector getKeys(CompanyStorage company) {
	return formula.getKeys(company);
    }

    // static getKeys - find the DynamicDataSet named ATTR, then get all
    // keys for it.
    static Vector getKeys(CompanyStorage company, String attr) {
	DynamicDataSet dds = find(attr);
	if (dds == null) { return null; }
	return dds.getKeys(company);
    }

    // ****************************************************************
    // Code for Groups of DynamicDataSets

    // A group of DynamicDataSets
    static class ddsGroup {
	String name;
	int priority;
	Hashtable defs;
	boolean active;

	// store a DDS into a group
	void store(String name, DynamicDataSet dds) {
	    defs.put(name, dds);
	}

	// find a DDS from a group
	DynamicDataSet find(String name) {
	    if (! active) { return null; }
	    return (DynamicDataSet) defs.get(name);
	}
    }

    // The list of all groups, active and non, in search order
    static Vector groups;

    // The current group, added to when defining a new DynamicDataSet
    static ddsGroup curGroup;

    // initialize group set and define the initial group
    static {
	groups = new Vector();
	newGroup("default", 0);
    };

    static ddsGroup findGroup(String name) {
	for (int i = 0; i < groups.size(); i++) {
	    ddsGroup curgrp = (ddsGroup) groups.elementAt(i);
	    if (name.equals(curgrp.name)) { return curgrp; }
	}
	return null;
    }
	
    // define a regular new group
    static public void newGroup(String name) {
	newGroup(name, 0);
    }

    // define a new group with specified priority
    static public void newGroup(String name, int priority) {
	
	ddsGroup grp = findGroup(name);

	if (grp != null) { return; }

	grp = new ddsGroup();
	grp.name = name;
	grp.priority = priority;
	grp.active = true;
	grp.defs = new Hashtable();

	curGroup = grp;
	// insert group into vector in priority order
	for (int i = 0; i < groups.size(); i++) {
	    ddsGroup curgrp = (ddsGroup) groups.elementAt(i);
	    if (priority >= curgrp.priority) {
		groups.insertElementAt(grp, i);
		return;
	    }
	}
	groups.addElement(grp);
    }

    // Set the default group to be the new group
    static public void useGroup(String name) {
	ddsGroup grp = findGroup(name);
	if (grp != null) {
	    curGroup = grp;
	}
    }

    // Return the current group name
    static public String curGroup() {
	return curGroup.name;
    }

    // delete a named group
    static public void delGroup(String name) {
	// remove group from Vector
	for (int i = 0; i < groups.size(); i++) {
	    ddsGroup curgrp = (ddsGroup) groups.elementAt(i);
	    if (name.equals(curgrp.name)) {
		groups.removeElementAt(i);
		return;
	    }
	}
    }

    // change a group priority
    static public void setPriority(String name, int priority) {
	ddsGroup grp = findGroup(name);
	if (grp != null) {
	    grp.priority = priority;
	}
    }

    // Activate a group
    static public boolean activate(String name) {
	return activate(name, true);
    }

    // Activate a group
    static public boolean activate(String name, boolean active) {
	ddsGroup grp = findGroup(name);
	if (grp != null) {
	    boolean oldflag = grp.active;
	    grp.active = active;
	    return oldflag;
	}
	return false;
    }

    static DynamicDataSet findDDS(String name) {
	// Debug.debug("Searching " + groups.size() + " groups for DDS " + name);
	// search active groups in vector for DDS of name
	for (int i = 0; i < groups.size(); i++) {
	    ddsGroup curgrp = (ddsGroup) groups.elementAt(i);
	    // Debug.debug("Searching group " + curgrp.name);
	    if (curgrp.active) {
		Object set = curgrp.defs.get(name);
		if (set != null) {
		    // Debug.debug("Found DDS.");
		    return (DynamicDataSet) set;
		}
	    }
	}
	return null;
    }
}
