
// Class Formula

// The Formula class implements a small calculations language, used to
// define formulas to be calculated.  Formulas are associated with data
// attributes (via the DynamicDataSet class), and perform mathematical
// operations based on other data attributes and constants.

// Examples:
//    Pre-Tax Profit Margin:
//                100 * "PreTax Income" / "Sales"
//    Last FY Quarter Data:
//                1 + (4 + ("Month of Latest Quarterly Data" / 3)
//                       - ("Fiscal Year End Month" / 3) - 1) % 4

// Formulas are written as standard mathematical equations,
//   e.g. x + 3, 2 * y, 4 * (a1 - a2)

// To reference another attribute, you can just put the attribute name into
// the formula, e.g. Sales / 1000

// If the attribute name contains spaces or other non-alphanumeric
// characters, then you must surround the attribute name with square
// brackets, e.g. [Earnings per Share], [Sales (in millions)], etc., though
// you can always use brackets even when not required.

// Functions can be put in formulas in standard C/C++/Java syntax,
// e.g. 2 * max([Long Term Debt], [Short Term Debt])

// Available operators/functions for Formulas are:
//  Binary operators
//    +		- addition
//    -		- subtraction
//    *		- multiplication
//    /		- division
//    %		- remainder
//
// These operators are also available in function syntax as:
//    plus, minus, times, dividedby, remainder
//    e.g. plus(3, 4, 7)
//
//  Binary (or N-ary) functions - these take two or more parameters
//    min	- minimum
//    max	- maximum
//    avg	- average

// Unary functions
//    abs	- absolute value

// Other
//    first - first available, e.g. first([Total Debt], [Long-Term Debt])
//              will return the total debt for the specified period, if
//              available.  If not available, then Long-Term Debt is
//              checked next.
//
//    prev - previous period's value for attribute;
//		examples:
//	            Sales - prev(Sales)
//				the increase in Sales from year to year
//		    100 * Sales / prev(Sales)
//				the % increase in Sales from year to year
//      prev can also take an optional second parameter to calculate
//	the no of periods difference, e.g.
//		100 * [Quarterly EPS] / prev([Quarterly EPS], 4)
//				calculates the year ago percentage change
//				in the quarterly earnings per share
//
//    latest - Most recent period's value for attribute;
//      This takes two parameters, the expression to evaluate and the
//      period frequency in which to evaluate it.
//        example: latest([Quarterly Sales], 4) returns the most recent
//            quarterly value for Quarterly Sales
//      Note: the most recent quarterly period is defined as the latest
//      quarterly period for which there is data defined for this company
//      for some attribute.
//
//    sequence - Generate a sequence of an expression for various dates,
//    to be used by another function.  This MUST be used internally
//    within a function call, as it doesn't return a real value.
//      This takes the following params:
//        - Formula to evaluate for each period
//        - Start period (relative to max period)
//        - End period (relative to max period)
//        - Increment
//    This function also removes non-numeric elements from the
//    resulting sequence.
//      examples:
//        plus(sequence([Quarterly Sales], -3, 0, 1))
//           - sums up the last four quarterly sales (going to the date
//           from three periods prior, counting by one)
//        avg(sequence([PE Ratio] * [Use PE Ratio Flag], 0, -4, -1));
//           - sums up the last five years of PE Ratio
//          e.g. above call is the same as:
//           avg(     [PE Ratio] * [Use PE Ratio Flag],
//               prev([PE Ratio] * [Use PE Ratio Flag], 1),
//               prev([PE Ratio] * [Use PE Ratio Flag], 2),
//               prev([PE Ratio] * [Use PE Ratio Flag], 3),
//               prev([PE Ratio] * [Use PE Ratio Flag], 4))
//          where values that don't calculate are removed
//          (i.e. if [Use PE Ratio Flag] is NULL or non-numeric for a
//          particular year).
//
//    select - Select an expression on the basis of an integer number.
//	Usage: select(<Selector>, <Expr>, <Expr>, ...)
// 	e.g. select(3, [Low Price by EPS], [Avg Low Price], [Severe Low Price],
//		    [Dividend Low Price], [Users Low Price])

//      would select the value for [Severe Low Price].  If the selector
//      number is out of range, then no value is returned.  This can be
//      useful with flags and sequences, e.g. you could calculate the
//      average of the previous five years of Low PEs, removing any
//      attributes that you don't want as part of the average with the
//      following expression:
//
//	avg(sequence(select([FLAG Use PE],[Low PE]), -4, 0, 1))
//
//      When [FLAG Use PE] is set to "1" for a particular year, then the
//      [Low PE] will be used, but when it's any other value, then nothing
//      will be returned and sequence will drop that item out.


package sptossg ;

import java.util.*;
import java.io.*;

public class Formula {

    // ****************************************************************
    // member variables

    // The string name of the function or operator
    String operatorStr;
    // the integer code for the operator
    int operator;
    // a stack of Formulas for the operator to be called on
    Stack params;

    // if the operator is FNFETCH, then the attribute to be retrieved
    String attribute;
    // if the operator is FNCONSTANT, then the value of the constant
    double value;

    // operator description variables
    // min & max no. of parameters allowed
    int minparams;
    int maxparams;
    // how many params to be evaluated
    int evalcount;
    // what is the return type of this function (numeric or sequence)
    int optype;

    // ****************************************************************
    // Constructors:

    // Binary operator constructor, e.g. A + B, mod(x, 4)
    public Formula (Formula left, String operator,  Formula right) {
	params = new Stack();
	params.push(right);
	params.push(left);
	initSelf(operator, params);
    }

    // Unary operator constructor, e.g. abs(a)
    public Formula (String operator, Formula left) {
	params = new Stack();
	params.push(left);
	initSelf(operator, params);
    }

    // N-ary operator constructor, e.g. max(a,b,c,d)
    public Formula (String operator,  Stack params) {
	initSelf(operator, params);
    }

    // actual initialization method for function operators
    private void initSelf(String operator, Stack params) {
	try {
	    Integer code = (Integer) OpsCodes.get(operator);
	    this.operator = code.intValue();
	    this.minparams = ((Integer) OpsMin.get(code)).intValue();
	    this.maxparams = ((Integer) OpsMax.get(code)).intValue();
	    this.evalcount = ((Integer) OpsEval.get(code)).intValue();
	    int optype = getOpType(code);
	    if (optype == OPboth) {
		if (params == null) {
		    optype = OPnumber;
		} else {
		    Stack nparams = (Stack) params.clone();
		    Formula f1 = (Formula) nparams.pop();
		    optype = f1.optype;
		}
	    }
	    this.optype = optype;

	} catch (Exception exc) {
	    System.out.println("Error initializing " + operator + "\n" + exc);
	    this.operator = -1;
	    return;
	}

	this.operatorStr = operator;
	this.params = params;
    }

    void describeSelf() {
	System.out.println("Formula: " + this);
	System.out.println("  operator: '" + operatorStr + "' (" + operator + ").");
	System.out.println("  param count: [" + minparams + "," + maxparams + "].");
	System.out.println("  evaluate: " + evalcount);
	System.out.println("  optype: " + optype);
    }

    // constructor for reference to a data attribute
    public Formula(String attr) {
	this.attribute = attr;
	this.operator = FNFETCH;
	this.operatorStr = "FETCH";
	this.optype = OPnumber;
	this.params = null;
    }

    // constructor for reference to a constant
    public Formula(double value) {
	this.value = value;
	this.operator = FNCONSTANT;
	this.operatorStr = "CONSTANT";
	this.optype = OPnumber;
	params = null;
    }

    // ****************************************************************

    // Does this formula return a sequence?
    boolean returnsSequence() {
	if (optype == OPsequence) { return true;
	} else { return false; }
    }

    // Does this formula return a number?
    boolean returnsNumber() {
	if (optype == OPnumber) { return true;
	} else { return false; }
    }

    // toString - Generate a readable test version of the formula, suitable
    // for re-parsing.

    public String toString() {
	// used for debugging
	boolean showsyntax = false;

	if (operatorStr.equals("CONSTANT")) {
	    // a reference to a Constant, e.g. 31.4
	    if (showsyntax) { return("const(" + value + ")");
	    } else { return("" + value); }
	} else if (operatorStr.equals("FETCH")) {
	    // A reference to an attribute retrieval
	    if (showsyntax) { return("get([" + attribute + "])");
	    } else { return("[" + attribute + "]"); }
	} else if (params.size() == 0) {
	    // A parameterless function
	    return(operatorStr + "()");
	} else if (params.size() == 1) {
	    // A single-parametered function
	    Formula param = (Formula) params.peek();
	    return(operatorStr + "(" + param + ")");
	} else if (operatorStr.length() == 1 && params.size() == 2) {
	    // An infix operator (e.g. 3 + 4)
	    Stack nparams = (Stack) params.clone();
	    Formula f1 = (Formula) nparams.pop();
	    Formula f2 = (Formula) nparams.pop();
	    return("(" + f1 + " " + operatorStr + " " + f2 + ")");
	} else {
	    // A multiple-parameter function call
	    Stack nparams = (Stack) params.clone();
	    String sep = "(";
	    String paramstr = "";
	    while (! nparams.empty()) {
		paramstr += sep + nparams.pop();
		sep = ", ";
	    }
	    return (operatorStr + paramstr + ")");
	}}

    // ****************************************************************
    // Static definitions for defining formula operators

    static Hashtable OpsCodes = new Hashtable();
    static Hashtable OpsMin = new Hashtable();
    static Hashtable OpsMax = new Hashtable();
    static Hashtable OpsEval = new Hashtable();

    static void DefineOp(String opstr, int code, int minargs,
			 int maxargs, int evalCount) {
	Integer icode = new Integer(code);

	if (opstr != null) {
	    OpsCodes.put(opstr, icode);
	}
	OpsMin.put(icode, new Integer(minargs));
	OpsMax.put(icode, new Integer(maxargs));
	OpsEval.put(icode, new Integer(evalCount));
    }

    static Hashtable OpsType = new Hashtable();

    static void DefineOpType(int code, int type) {
	Integer icode = new Integer(code);
	Integer itype = new Integer(type);
	OpsType.put(icode, itype);
    }

    static int getOpType(Integer icode) {
	Object itm = OpsType.get(icode);

	// numeric is the default
	if (itm == null) { return OPnumber; }
	return ((Integer) itm).intValue();
    }

    static public final int OPnumber = 1;
    static public final int OPsequence = 2;
    static public final int OPboth = 11;


    // Define integer constants for available functions and internal operators

    // internal operators
    static public final int FNFETCH = 1;
    static public final int FNCONSTANT = 2;

    // date modifying operators
    static public final int FNPREV = 11;
    static public final int FNLATEST = 12;

    // special operators - evaluation of params is handled differently
    static public final int FNFIRST = 21;
    static public final int FNSEQUENCE = 22;
    static public final int FNSELECT = 23;
    static public final int FNYEAR = 24;
    static public final int FNPERIOD = 25;

    // regular operators (unary, binary, N-ary)
    static public final int FNABS = 101;
    static public final int FNPLUS = 103;
    static public final int FNMINUS = 104;
    static public final int FNTIMES = 105;
    static public final int FNDIVIDE = 106;
    static public final int FNREMAINDER = 107;
    static public final int FNMIN = 108;
    static public final int FNMAX = 109;
    static public final int FNFLOOR = 110;
    static public final int FNAVG = 111;
    static public final int FNPOW = 112;

    // Define all built-in functions and internal operators
    static {
	// DefineOp - takes five params:
	//    function name - string name for parsing
	//    function code - integer representation
	//    min and max no of params
	//    how many params to evaluate - usually the max, except
	//	for special functions

	// **************** Internal Operators ****************

	// Usage: none explicit (internally generated)
	//   e.g. [Sales]
	DefineOp(null, FNFETCH, 0, 0, 0);

	// Usage: none explicit (internally generated)
	//   e.g. 3.5
	DefineOp(null, FNCONSTANT, 0, 0, 0);

	// **************** Date Modifying Functions ****************

	// Usage: prev(<Expr>) or prev(<Expr>, <no of periods>)
	//  e.g. prev(EPS), prev([quarterly sales], 4)
	DefineOp("prev", FNPREV, 1, 2, 0);

	// Usage: latest(<Expr>, <frequency>)
	//  e.g. latest(EPS, 1)
	DefineOp("latest", FNLATEST, 2, 2, 0);
	DefineOpType(FNLATEST, OPboth);

	// **************** Special Evaluation Functions ****************

	// Usage: first(<Expr>, <Expr>, ...)
	//  e.g. first([Shares Outstanding], [Calculated Shares])
	DefineOp("first", FNFIRST, 2, 100, 0);

	// Usage: sequence(<Expr>, <start offset>, <end offset>, <increment>)
	//  e.g. avg(sequence([PE Ratio] * [Use PE Ratio Flag], 0, 4, -1));
	//  e.g. above call is the same as:
	//   avg(     [PE Ratio] * [Use PE Ratio Flag],
	//       prev([PE Ratio] * [Use PE Ratio Flag], 1),
	//       prev([PE Ratio] * [Use PE Ratio Flag], 2),
	//       prev([PE Ratio] * [Use PE Ratio Flag], 3),
	//       prev([PE Ratio] * [Use PE Ratio Flag], 4))
	//  where values that don't calculate are removed
	// (e.g. if [Use PE Ratio Flag] is NULL for a particular year).
	DefineOp("sequence", FNSEQUENCE, 4, 4, 0);
	DefineOpType(FNSEQUENCE, OPsequence);

	// Usage: select(<selector count>, <Expr>, <Expr>, ...)
	//  e.g.
	// select([low price indicator], [Low Price by EPS], [Avg Low Price],
	//        [Severe Low Price], [Dividend Low Price], [Users Low Price])
	// where [low price indicator] is a number from one to five
	DefineOp("select", FNSELECT, 2, 100, 1);

	// Usage: year(), period()
	//  Most useful wrapped in latest, e.g. (latest(year(), 1)), (latest(period(), 4))
	DefineOp("year", FNYEAR, 0, 0, 0);
	DefineOp("period", FNPERIOD, 0, 0, 0);

	// **************** Regular Operators & Functions ****************

	// Usage: abs(<Expr>)
	//  e.g. abs(-17.4)
	DefineOp("abs", FNABS, 1, 1, 1);

	// Usage: <Expr> + <Expr>, plus(<Expr>, <Expr>, ...)
	//  e.g. 23 + 11, plus(23,11)
	DefineOp("+", FNPLUS, 1, 100, 100);
	DefineOp("plus", FNPLUS, 1, 100, 100);

	// Usage: <Expr> - <Expr>, minus(<Expr>, <Expr>, ...)
	//  e.g. [Growth Rate] - 1, minus([Growth Rate], 1)
	DefineOp("-", FNMINUS, 1, 100, 100);
	DefineOp("minus", FNMINUS, 1, 100, 100);

	// Usage: <Expr> * <Expr>, times(<Expr>, <Expr>, ...)
	//  e.g. [Tax] * [Profits], times([Tax], [Profits])
	DefineOp("*", FNTIMES, 1, 100, 100);
	DefineOp("times", FNTIMES, 1, 100, 100);

	// Usage: <Expr> / <Expr>, dividedby(<Expr>, <Expr>, ...)
	//  e.g. [Price] / [Earnings], dividedby([Price], [EARNINGS]);
	DefineOp("/", FNDIVIDE, 1, 100, 100);
	DefineOp("dividedby", FNDIVIDE, 1, 100, 100);

	// Usage: <Expr> % <Expr>, remainder(<Expr>, <Expr>, ...)
	//  e.g. [months] % 12
	DefineOp("%", FNREMAINDER, 2, 2, 2);
	DefineOp("remainder", FNREMAINDER, 2, 2, 2);

	// Usage: min(<Expr>, <Expr>, ...)
	//  e.g. min([Net Profit], [Taxes], [Expenditures])
	DefineOp("min", FNMIN, 2, 100, 100);

	// Usage: max(<Expr>, <Expr>, ...)
	//  e.g. max([Net Profit], [Taxes], [Expenditures])
	DefineOp("max", FNMAX, 2, 100, 100);

	// Usage: floor(<Expr>)
	//  e.g. floor([Sales] / 1000)
	DefineOp("floor", FNFLOOR, 1, 1, 1);

	// Usage: avg(<Expr>, <Expr>, ...)
	//  e.g. avg([Sales], prev([Sales]))
	DefineOp("avg", FNAVG, 1, 100, 100);

	// Usage: pow(<Expr>, <Expr>)
	//  e.g. . pow(1 + [Rate], [Years])
	DefineOp("pow", FNPOW, 2, 2, 2);

    }

    // ****************************************************************
    // code to parse string representations of a formula
    //  e.g. "Sales / 1000", "100 * [PreTax Inc] / Sales",
    //    "max(Sales - Expenses, 0)"

    // Formula scanner - breaks up a formula string into a Stack of tokens
    static Stack tokenize(String formulastr) {
	StringTokenizer st = new StringTokenizer(formulastr, " ", true);
	Stack tokens = new Stack();
	while (st.hasMoreTokens()) {
	    String token = nextToken(st);
	    tokens.addElement(token);
	}

        // reverse tokens so that first tokens are at top of Stack and add
        // an end-marker.
	Stack result = new Stack();
	result.push("EOF");
	while (! tokens.empty()) {
	    result.push(tokens.pop());
	}
	return result;
    }

    // Scan the string to return the next token.  A token is:
    //   An operator - A special character - e.g. "+", "-", "*", "/", "%", etc.
    //   A number - a sequence of digits
    //   An Attribute - a sequence of characters, starting with an alphabetic.

    //      Note: if an attribute has any spaces or special characters
    //      contained within, then it can be surrounded by square brackets
    //      to denote the entire unit,
    //      e.g. [Net Profit], [Sales (in millions)].  The scanner will
    //      strip off the brackets before returning the token.
    static String nextToken(StringTokenizer st) {
	String sepchars = "!@#$%^&*()-=+[{]};:',/<>? \t";
	String token = null;
	while (token == null) {
	    token = st.nextToken(sepchars);
	    if (token.trim().length() == 0) {
		token = null;
	    } else {
		if (token.equals("[")) {
		    token = st.nextToken("]");
		    st.nextToken(sepchars); // skip the ']'
		}
	    }
	}
	return token;
    }

    // ****************************************************************
    // parse(<FORMULA STRING>) - A recursive descent parser for formulas.
    // Formulas can be described by the following grammar:

    // E           -> E + E | E - E | E + E | E / E | E % E | ( E ) |
    //		      ATOM | NUMBER | FN-CALL
    // FN-CALL     -> ATOM ( PARAM-LIST )
    // PARAM-LIST  -> E | E , PARAM-LIST

    // However, recursive descent parsers require an LL(1) grammar.  The following
    // is the LL(1) grammar used for the formula language:

    // 	S	E1 EOF

    // 	E1	E2 E1Rest

    // 	E1Rest	"+" E2 E1Rest
    // 		"-" E2 E1Rest
    // 		<null>

    // 	E2	ATOM E2Rest

    // 	E2Rest	"*" ATOM E2Rest
    // 		"/" ATOM E2Rest
    // 		"%" ATOM E2Rest
    // 		<null>

    // 	ATOM	NUMBER
    // 		"(" E1 ")"
    // 		"-" ATOM
    // 		OPERAND "(" PARAMS
    // 		OPERAND

    // 	PARAMS	")"
    // 		E1 ")"
    // 		E1 "," PARAMS ")"

    // The parser for this grammar is implemented as a recursive descent
    // parser.  All non-terminals in the grammar are parsed via their own
    // parsing routines, named parse_<NON-TERMINAL>.  Each parse routine
    // takes the stack of tokens and returns a Formula representing that
    // sub-expression.  In the cases where a non-terminal does not
    // represent a complete sub-expression (E1Rest & E2Rest), the parse
    // routine takes an additional parameter which completes the formula.


    public static Formula parse(String formulastr) {
	Stack toks = tokenize(formulastr);
	Formula f;

	execution = new Stack();
	try {
	    f = parse_S(toks);
	} catch (Exception exc) {
	    System.out.println("Error on parse: " + exc);
	    dumpstack(execution,"Execution Stack");
	    return null;
	}
	if (f == null) { return null; }

	return f;
    }

    static Stack execution;
    static void enter(String name) {
	execution.push(name);
    }

    static void exit(String name) {
	String old = (String) execution.pop();
    }

    static void parse_error(Stack tokens, String node) {
	System.out.println("Parse Error: " + node);
    }

    // Parser for S
    // 	S	E1 EOF
    static Formula parse_S(Stack tokens) {
	enter("parse_S");
	Formula f = parse_E1(tokens);
	if (f == null) { return null; }
	String token = (String) tokens.pop();
	if (token.equals("EOF") && tokens.empty()) {
	    exit("parse_S");
	    return f;
	}
	parse_error(tokens, "parse_S");
	return null;
    }

    // Parser for E1
    // 	E1	E2 E1Rest
    static Formula parse_E1(Stack tokens) {
	enter("parse_E1");
	Formula f1 = parse_E2(tokens);
	if (f1 == null) { return null; }
	Formula f = parse_E1Rest(tokens, f1);
	if (f != null) {
	    exit("parse_E1");
	    return f;
	}
	parse_error(tokens, "parse_E1");
	return null;
    }

    // Parser for E1Rest
    // 	E1Rest	"+" E2 E1Rest
    // 		"-" E2 E1Rest
    // 		<null>
    static Formula parse_E1Rest(Stack tokens, Formula f1) {
	enter("parse_E1Rest");
	String op = (String) tokens.pop();
	Formula f2 = null;
	Formula f3 = null;

	if (op.equals("+") || op.equals("-")) {
	    f2 = parse_E2(tokens);
	}

	if (f2 == null) {
	    tokens.push(op);
	    exit("parse_E1Rest");
	    return f1;
	}

	f3 = parse_E1Rest(tokens, new Formula(f1, op, f2));
	exit("parse_E1Rest");
	return f3;
    }

    // Parser for E2
    // 	E2	ATOM E2Rest
    static Formula parse_E2(Stack tokens) {
	enter("parse_E2");
	Formula f1 = parse_ATOM(tokens);
	if (f1 == null) { return null; }
	Formula f2 = parse_E2Rest(tokens, f1);
	exit("parse_E2");
	return f2;
    }

    // Parser for E2Rest
    // 	E2Rest	"*" ATOM E2Rest
    // 		"/" ATOM E2Rest
    // 		"%" ATOM E2Rest
    // 		<null>
    static Formula parse_E2Rest(Stack tokens, Formula f1) {
	enter("parse_E2Rest");
	String op = (String) tokens.pop();
	Formula f2 = null;
	Formula f3 = null;

	if (op.equals("*") || op.equals("/") || op.equals("%")) {
	    f2 = parse_ATOM(tokens);
	}

	if (f2 == null) {
	    tokens.push(op);
	    exit("parse_E2Rest");
	    return f1;
	}

	f3 = parse_E2Rest(tokens, new Formula(f1, op, f2));
	exit("parse_E2Rest");
	return f3;
    }

    // 	ATOM	NUMBER
    // 		"(" E1 ")"
    // 		"-" ATOM
    // 		OPERAND "(" PARAMS
    // 		OPERAND
    static Formula parse_ATOM(Stack tokens) {
	enter("parse_ATOM");
	Formula f1;
	String token = (String) tokens.pop();

	if (tokenIsNumber(token)) {
	    f1 = new Formula((new Double(token)).doubleValue());
	    exit("parse_ATOM");
	    return f1;
	}
	if (token.equals("(")) {
	    f1 = parse_E1(tokens);
	    token = (String) tokens.pop();
	    if (token.equals(")")) {
		exit("parse_ATOM");
		return f1;
	    }
	    parse_error(tokens, "parse_ATOM-1");
	    return null;
	}
	if (token.equals("-")) {
	    f1 = parse_ATOM(tokens);
	    Formula f2 = new Formula("-", f1);
	    exit("parse_ATOM");
	    return f2;
	}
	if (tokenIsOperand(token)) {
	    String next = (String) tokens.pop();
	    if (next.equals("(")) {
		Stack params = parse_PARAMS(tokens);

		// verify function in function call
		Integer code = (Integer) OpsCodes.get(token);

		if (code == null) {
		    parse_error(tokens, "Function error: " +
				"unknown function " + token);
		    return null;
		}

		// verify parameter count for function
		int minparams;
		int maxparams;
		try {
		    minparams = ((Integer) OpsMin.get(code)).intValue();
		    maxparams = ((Integer) OpsMax.get(code)).intValue();
		} catch (Exception exc) {
		    System.out.println("Error: " + exc);
		    return null;
		}

		int size = params.size();
		if (size < minparams || size > maxparams) {
		    parse_error(tokens, "Function error: " +
				size + " is invalid number of params" +
				" [" + minparams + "," + maxparams + "] " +
				", function "
				+ token);
		    return null;
		}

		f1 = new Formula(token, params);
		exit("parse_ATOM");
		return f1;
	    }
	    tokens.push(next);

	    f1 = new Formula(token);
	    exit("parse_ATOM");
	    return f1;
	}
	parse_error(tokens, "parse_ATOM-2");
	return null;
    }

    // Parser for PARAMS
    // 	PARAMS	")"
    // 		E1 ")"
    // 		E1 "," PARAMS ")"
    //

    // Note: this is the one inconsistent parse routine in that it doesn't
    // return a Formula like all the others, but instead returns a Stack of
    // Formulas.
    static Stack parse_PARAMS(Stack tokens) {
	enter("parse_PARAMS");
	Stack formulas;

	String sep = (String) tokens.pop();
	if (sep.equals(")")) {
	    exit("parse_PARAMS");
	    return new Stack();
	} else {
	    tokens.push(sep);
	}

	Formula f1 = parse_E1(tokens);
	if (f1 == null) {
	    parse_error(tokens, "parse_PARAMS-1");
	    return null;
	}
	sep = (String) tokens.pop();
	if (sep.equals(")")) {
	    formulas = new Stack();
	} else if (sep.equals(",")) {
	    formulas = parse_PARAMS(tokens);
	} else {
	    parse_error(tokens, "parse_PARAMS-2");
	    return null;
	}
	formulas.push(f1);
	exit("parse_PARAMS");
	return formulas;
    }

    // a token is numeric if the first character is numeric, though all
    // characters should be numeric
    static boolean tokenIsNumber(String token) {
	char c1 = token.charAt(0);
	if (c1 >= '0' && c1 <= '9') { return true; }
	if (c1 == '.') { return true; }
	return false;
    }

    // a token is an operand if the first character is alphabetic.
    // Subsequent characters can be of any type.
    static boolean tokenIsOperand(String token) {
	char c1 = token.charAt(0);
	if (c1 >= 'a' && c1 <= 'z') { return true; }
	if (c1 >= 'A' && c1 <= 'Z') { return true; }
	return false;
    }

    // an operator is generally a single character (not enforced), and not
    // a number or operand.
    static boolean tokenIsOperator (String token) {
	return ((! tokenIsOperand(token)) && (! tokenIsNumber(token)));
    }

    // debugging routine to dump a stack
    static void dumpstack(Stack toks, String title) {
	System.out.println(title);

	// clone the stack so that we don't modify it.
	Stack s1 = (Stack) toks.clone();
	int i = 0;

	while (! s1.empty()) {
	    String val = (String) s1.pop();
	    System.out.println(i++ + ": " + val);
	}
    }

    // ****************************************************************
    // Utility routines

    // Merge two sorted vectors of Periods into a new sorted vector
    static public Vector merge(Vector v1, Vector v2) {
	if (v1 == null) { return v2; }
	if (v2 == null) { return v1; }

	Vector result = new Vector();

	while ((! v1.isEmpty()) && (! v2.isEmpty())) {
	    Period p1 = (Period) v1.firstElement();
	    Period p2 = (Period) v2.firstElement();

	    // merge vectors, eliminating duplicates
	    if (p1.equals(p2)) {
		result.addElement(p1);
		v1.removeElementAt(0);
		v2.removeElementAt(0);
	    } else if (p1.less(p2)) {
		result.addElement(p1);
		v1.removeElementAt(0);
	    } else {
		result.addElement(p2);
		v2.removeElementAt(0);
	    }
	}

        // One vector was cleaned out; copy remaining elements from the
        // other vector
	Vector v4 = v1;
	if (v1.isEmpty()) { v4 = v2; }

	while (! v4.isEmpty()) {
	    result.addElement(v4.firstElement());
	    v4.removeElementAt(0);
	}

	return result;
    }

    // get all Period keys for a Formula; This is the *potential* domain
    // for the Formula, calculated as the union of the domain of periods
    // for each attribute referenced within.  Because this is a union, and
    // due to the vagaries of the formula execution, some of these keys may
    // actually not be part of the domain of this Formula.
    public Vector getKeys(CompanyStorage company) {

	if (operator == FNFETCH) { return company.getKeys(attribute); }

	if (params == null) { return null; }

	Vector keys = null;

	Stack nparams = (Stack) params.clone();
	while (! nparams.empty()) {
	    Formula f1 = (Formula) nparams.pop();
	    keys = merge(keys, f1.getKeys(company));
	}
	return keys;
    }

    boolean between(int val, int low, int high) {
	if (val < low) { return false; }
	if (val > high) { return false; }
	return true;
    }

    boolean isNum(double val) {
	if (Double.isNaN(val) || Double.isInfinite(val)) { return false; }
	return true;
    }

    // Parse a string as a number, returning a double
    //  This takes the following forms of numbers,
    //    - decimal form, e.g. "37", "3.14159", "2,300"
    //    - fractional form, e.g. "37 5/8", "22 3/4"
    //    - percentages, e.g. "37%"
    static double parseNumber(String numstr) {
	// first remove commas, e.g. "14,273,912"
	int comma = numstr.indexOf(',');
	while (comma >= 0) {
	    numstr = numstr.substring(0, comma) + numstr.substring(comma+1);
	    comma = numstr.indexOf(',');
	}

	double factor = 1;

	String smain;
	String snumerator;
	String sdenominator;

	// check if format is a percentage
	int pctg = numstr.indexOf('%');
	if (pctg != -1) {
	    numstr = numstr.substring(0, pctg);
	    factor = 100;
	}

	// check if format is fractional
	int space = numstr.indexOf(' ');
	int slash = numstr.indexOf('/');

	if (space == -1) {
	    smain = numstr;
	    snumerator = "0";
	    sdenominator = "1";
	} else {
	    // If no '/' or precedes space
	    if (slash <= space) { return Double.NaN; }

	    smain = numstr.substring(0, space);
	    snumerator = numstr.substring(space+1, slash);
	    sdenominator = numstr.substring(slash+1);
	}
	try {
	    double main = (new Double(smain)).doubleValue();
	    double numerator = (new Double(snumerator)).doubleValue();
	    double denominator = (new Double(sdenominator)).doubleValue();

	    double val = main + (numerator / denominator);
	    val = val / factor;
	    return val;
	} catch (Exception exc) {
	    return Double.NaN;
	}
    }

	

    // ****************************************************************
    // Formula evaluation routines

    // eval - an interpretive formula evaluator

    //  This takes a string definition for a formula and a company, and
    //  evaluates the formula in the context of that company.  This throws
    //  the exception java.text.ParseException if the formula definition
    //  string doesn't parse correctly.

    public static double eval(String def, CompanyStorage company)
    throws java.text.ParseException {
	return eval(def, company, null);
    }

    public static double eval(String def, CompanyStorage company, Period date)
	throws java.text.ParseException {
	Formula f = parse(def);
	if (f != null) {
	    return f.evaluate(company, date);
	} else {
	    throw new java.text.ParseException("Illegal function '" + def + "'",
					       0);
	}
    }

    // seqEvaluate - evaluate a sequence function, returning a Vector of
    //  the items.
    public Enumeration seqEvaluate(CompanyStorage company, Period date) {
	Vector itms = new Vector();

	// clone the parameter stack so that we can process it without
	// modifying it.
	Stack nparams = (Stack) params.clone();

	switch (operator) {
	case FNSEQUENCE:
	    {
		//  This takes the following params:
		//    - Formula to evaluate for each period
		//    - Start period (relative to max period)
		//    - End period (relative to max period)
		//    - Increment
		if (params.size() != 4) {
		    System.out.println("Formula error: illegal number of parameters: "
				       + this);
		    break;
		}

		Formula f1 = (Formula) nparams.pop();
		Formula f2 = (Formula) nparams.pop();
		Formula f3 = (Formula) nparams.pop();
		Formula f4 = (Formula) nparams.pop();

		int start = (int) f2.evaluate(company, date);
		int end = (int) f3.evaluate(company, date);
		int incr = (int) f4.evaluate(company, date);

		if (incr == 0) { break; }

		Period curper = date.adjust(start);

		int cur = start;

		while (between(cur, start, end) || between(cur, end, start)) {
		    if (Math.abs(start) > 10) { break ; }
		    double val = f1.evaluate(company, curper);
		    if (isNum(val)) {
			itms.addElement(new Double(val));
		    }
		    curper = curper.adjust(incr);
		    cur += incr;
		}

		break;
	    }
	case FNLATEST: {
	    Formula f1 = (Formula) nparams.pop();

	    if (! f1.returnsSequence()) {
		System.out.println("LATEST can't take the return value for " + f1);
		return itms.elements();
	    }

	    int freq;

	    Formula f2 = (Formula) nparams.pop();
	    if (! f2.returnsNumber()) {
		System.out.println("LATEST can't take the return value for " + f2);
		return itms.elements();
	    }

	    freq = (int) f2.evaluate(company, date);
	    Period latest = company.getMaxPeriod(freq);
	    return f1.seqEvaluate(company, latest);
	}

	default:
	    System.out.println("Unknown sequence operator " + operator);
	}

	return itms.elements();
    }

    // evaluation for non-date formulas
    public double evaluate(CompanyStorage company) {
	return evaluate(company, null);
    }

    // evaluate a formula within the context of a company, for a particular
    // date.  Method returns a double as a value, or the constant
    // Double.NaN if the formula doesn't have a value for that date.
    //

    // This routine contains all the code implementing every standard
    // built-in function (only sequence functions are elsewhere), though
    // they should probably be moved out into new class definitions.

    public double evaluate(CompanyStorage company, Period date) {

	if (company == null) {
	    System.out.println("Error: no company passed for formula " +
			       this.toString());
	    return Double.NaN;
	}

	double args[] = new double[0];
	int argctr = 0;

	int evalcount = this.evalcount;

	// evaluate parameter stack
	if (params != null && evalcount > 0) {
	    int arrayelts = evalcount;
	    args = new double[arrayelts];
	    Stack nparams = (Stack) params.clone();
	    while (! nparams.empty() && evalcount > 0) {
		evalcount--;
		Formula f1 = (Formula) nparams.pop();
		if (f1.returnsNumber()) {
		    double val = f1.evaluate(company, date);
		    if (! isNum(val)) { return Double.NaN; }
		    if (argctr >= arrayelts) {
			System.out.println("Error: Too many elements to " +
					   "evaluate on " + this);
			return Double.NaN;
		    }
		    args[argctr++] = val;
		} else if (f1.returnsSequence()) {
		    Enumeration valseq = f1.seqEvaluate(company, date);
		    while (valseq.hasMoreElements()) {
			if (argctr >= arrayelts) {
			    System.out.println("Error: Too many elements to " +
					       "evaluate on " + this);
			    return Double.NaN;
			}
			Double elt = (Double) valseq.nextElement();
			double val = elt.doubleValue();
			args[argctr++] = val;
		    }
		} else {
		    System.out.println("Unknown return type for formula " + f1);
		    return Double.NaN;
		}
	    }
	}

	// build a copy of parameters for fns that don't evaluate all args
	Stack nparams = null;
	if (params != null && evalcount < maxparams) {
	    nparams = (Stack) params.clone();
	}

	double result;
	switch (operator) {
	case FNFETCH:
	    String valstr = null;

	    // if date given, do a date-specific retrieval
	    if (date != null) {
		valstr = company.get(attribute, date);
	    }

            // if no date given or no value found for date, then see if
            // it's a non-date-specific attribute.
	    if (valstr == null) {
		valstr = company.get(attribute);
	    }

	    // if no value found, signal error
	    if (valstr == null) { return Double.NaN; }

	    result = parseNumber(valstr);
	    break;
	case FNCONSTANT:
	    result = value;
	    break;
	case FNFIRST:
	    result = Double.NaN;
	    while (! nparams.empty()) {
		Formula f1 = (Formula) nparams.pop();
		if (f1.returnsNumber()) {
		    double val = f1.evaluate(company, date);
		    if (isNum(val)) {
			result = val;
			break;
		    }
		}
	    }
	    break;
	case FNSELECT: {
	    int selector = (int) args[0];

	    if (selector < 1) { return Double.NaN; }
	    // skip selector count
	    Formula f1 = (Formula) nparams.pop();
	    while (! nparams.empty()) {
		f1 = (Formula) nparams.pop();
		selector--;
		// found selected item
		if (selector == 0) {
		    double val = f1.evaluate(company, date);
		    if (isNum(val)) { return val; }
		    return Double.NaN;
		}
	    }
	    // not enough items, return nothing
	    result = Double.NaN;
	    break;
	}
	case FNYEAR:
	case FNPERIOD:
	    // case FNFREQUENCY:
	    {
		if (date == null) {
		    System.out.println("Error: no date passed to date function year/period. ");
		    return Double.NaN;
		}
		int val = -1;
		if (operator == FNYEAR) {
		    val = date.year();
		} else if (operator == FNPERIOD) {
		    val = date.period();
		}
		if (val < 0) {
		    return Double.NaN;
		}
		result = val;
		break;
	    }
	case FNPREV: {
	    Formula f1 = (Formula) nparams.pop();

	    if (! f1.returnsNumber()) {
		System.out.println("PREV can't take the return value for " + f1);
		return Double.NaN;
	    }

	    int offset;

	    if (! nparams.empty()) {
		Formula f2 = (Formula) nparams.pop();
		if (! f2.returnsNumber()) {
		    System.out.println("PREV can't take the return value for " + f2);
		    return Double.NaN;
		}
		offset = (int) f2.evaluate(company, date);
	    } else {
		offset = 1;
	    }

	    Period newdate = date.adjust(- offset);

	    result = f1.evaluate(company, newdate);
	    break;
	}
	case FNLATEST: {
	    Formula f1 = (Formula) nparams.pop();

	    if (! f1.returnsNumber()) {
		System.out.println("LATEST can't take the return value for " + f1);
		return Double.NaN;
	    }

	    int freq;

	    Formula f2 = (Formula) nparams.pop();
	    if (! f2.returnsNumber()) {
		System.out.println("LATEST can't take the return value for " + f2);
		return Double.NaN;
	    }

	    freq = (int) f2.evaluate(company, date);
	    Period latest = company.getMaxPeriod(freq);
	    result = f1.evaluate(company, latest);
	    break;
	}
	case FNABS:
	    result = Math.abs(args[0]);
	    break;
	case FNPLUS:
	    result = args[0];
	    for (int i = 1; i < argctr; i++) {
		result += args[i];
	    }
	    break;
	case FNMINUS:
	    if (argctr == 1) {
		result = 0 - args[0];
	    } else {
		result = args[0];
		for (int i = 1; i < argctr; i++) {
		    result -= args[i];
		}
	    }
	    break;
	case FNTIMES:
	    result = args[0];
	    for (int i = 1; i < argctr; i++) {
		result *= args[i];
	    }
	    break;
	case FNDIVIDE:
	    if (argctr == 1) {
		result = 1 / args[0];
	    } else {
		result = args[0];
		for (int i = 1; i < argctr; i++) {
		    result /= args[i];
		}
	    }
	    break;
	case FNREMAINDER:
	    result = args[0] % args[1];
	    break;
	case FNMIN:
	    result = Math.min(args[0], args[1]);
	    break;
	case FNMAX:
	    result = Math.max(args[0], args[1]);
	    break;
	case FNFLOOR:
	    int floorval = (int) args[0];
	    result = floorval;
	    break;
	case FNAVG:
	    double total = 0;
	    for (int i = 0; i < argctr; i++) {
		total += args[i];
	    }
	    result = total / argctr;
	    break;
	case FNPOW:
	    result = Math.pow(args[0], args[1]);
	    break;
	default:
	    System.out.println("Unknown numeric operator " + operator);
	    return Double.NaN;
	}

	return result;
    }
}
