package sptossg;

import java.io.*;
import gnu.regexp.*;

/*
 * Notes:
 *
 * <ul>
 * <li>Fiscal year ending should be converted to a Period.
 * <li>Should report date also be converted?
 * </ul>
 */

public class SPReportParser
{
	// determines whether debug messages are printed
	static final boolean DEBUG = true;
	
	/** name of file being parsed */
	String fileName;
	
	/** input file */
	FileReader theFile;
	BufferedReader inFile;
 
	/** error string if one occurred */
	String parseErr;
	
	/** data from being currently processed -- all lines */
	StringBuffer lines;
	
	/**
	 * Construct parser object using
	 * name of file to parse.
	 */
	
	public SPReportParser(String fileName) throws FileNotFoundException
	{
		this.fileName = fileName;
		// inFile = new FileReader(fileName);
		theFile = new FileReader(fileName);
		inFile = new BufferedReader(theFile);
		parseErr = "No error!";
	}

	/*
	 * prints info about blocks -- used by parse() during debugging
	 */
	static void block(String name, String data)
	{
		System.out.println("----------------------------------------");
		System.out.print("block name: ");
		System.out.println(name);
		System.out.println("----------------------------------------");
		System.out.println(data);
		System.out.println();
	}

	/**
	 * main() for debugging
	 */
	public static void main(String args[]) throws java.io.FileNotFoundException
	{
		for (int i = 0; i < args.length; i++) {
			SPReportParser parser = new SPReportParser(args[i]);
			CompanyInfo data = new CompanyInfo();
			if (!parser.parse(data)) {
				System.err.println("parse() failed with the message: " 
								   + parser.parseError());
			}
			data.dump();
		}
	}

	/**
	 * Parse everything to set values in a
	 * CompanyInfo object.
	 */

	public boolean parse(CompanyInfo data)
	{
		try	{
			lines = new StringBuffer();
			while (inFile.ready()) {
				lines.append(inFile.readLine());
				lines.append('\n');
			}

			// header section
			String header;
			SectionParser secParser = new SectionParser(lines.toString());
			REMatch match = secParser.getSection("<table.*?>(.+?s_plogo.+?)</table>");
			if (match != null) {
				header = match.toString(1);
				processHeader(header, data);
			}
			else {
				parseErr = "Header section not found";
				return false;
			}

			// industry section
			String industry;
			match = secParser.getSection("<table.*?>(.+?Industry:.+?)</table>");
			if (match != null) {
				industry = match.toString(1);
				processIndustry(industry, data);
			}
			else {
				parseErr = "Industry section not found";
				return false;
			}

			// price section
			String price;
			match = secParser.getSection("<table.*?>(.+?Recent Price.+?)</table>");
			if (match != null) {
				price = match.toString(1);
				processPrice(price, data);
			}
			else {
				parseErr = "Price section not found";
				return false;
			}

			// evaluation section
			String evaluation;
			match = secParser.getSection("<table.*?>(.+?Quantitative Evaluation.+?)</table>");
			if (match != null) {
				evaluation = match.toString(1);
				processEvaluation(evaluation, data);
			}
			else {
				parseErr = "Evaluation section not found";
				return false;
			}

			// overview section
			String overview;
			match = secParser.getSection("Overview.+?</multicol>");
			if (match != null) {
				overview = match.toString();
				// no parsing of overview for now - it's all text
			}
			else {
				parseErr = "Overview section not found";
				return false;
			}

			// valuation section
			String valuation;
			match = secParser.getSection("Valuation.+?</multicol>");
			if (match != null) {
				valuation = match.toString();
				// no parsing of valuation for now - it's all text
			}
			else {
				parseErr = "Valuation section not found";
				return false;
			}

			// key stock statistics section
			String keyStats;
			match = secParser.getSection("<table.*?>(.+?Beta.+?)</table>");
			if (match != null) {
				keyStats = match.toString(1);
				processKeyStats(keyStats, data);
			}
			else {
				parseErr = "Key stock statistics section not found";
				return false;
			}

			// get "value of $10,000..." value and set it
			match = secParser.getSection("Value of \\$10,000.+?:(.+?)<");
			if (match != null) {
				data.set(match.toString(1), "value of $10,000");
			}
			else {
				parseErr = "Value of $10,000 section not found";
				return false;
			}
		
			// get fiscal year ending value and set it
			match = secParser.getSection("Fiscal Year Ending (.+?)<");
			if (match != null) {
				data.set(match.toString(1), "fy ending");
			}
			else {
				parseErr = "FY ending section not found";
				return false;
			}

			// revenues section
			String revenues;
			match = secParser.getSection("Revenue.+?<table.*?>(.+?)</table>");
			if (match != null) {
				revenues = match.toString(1);
				processRevenues(revenues, data);
			}
			else {
				parseErr = "Revenues section not found";
				return false;
			}

			// eps section
			String eps;
			match = secParser.getSection("Earnings Per Share.+?<table.*?>(.+?)</table>");
			if (match != null) {
				eps = match.toString(1);
				processEPS(eps, data);
			}
			else {
				parseErr = "Earnings per share section not found";
				return false;
			}

			// get dividends paid since date if it shows up
			match = secParser.getSection("Dividends have been paid since\\s+(\\d+)");
			if (match != null)
				data.set(match.toString(1), "dividends paid since");
			// not really an error if we don't find it, so no "else" here
		
			// get business summary and set it
			match = secParser.getSection("Business Summary.+?<multicol.+?>.+?<font.+?>(.+?)</font>");
			if (match != null) {
				data.set(match.toString(1).replace('\n', ' '), "business summary");
			} // not really an error if we don't find it, so no "else" here

			// per share data section
			String perShare;
			match = secParser.getSection("Per Share Data.+?<table.*?>(.+?)</table>");
			if (match != null) {
				perShare = match.toString(1);
				processPerShareData(perShare, data);
			}
			else {
				parseErr = "Per share data section not found";
				return false;
			}

			// income statement section
			String income;
			match = secParser.getSection("(Income Statement.+?)<table.*?>(.+?)</table>");
			if (match != null) {
				// get scaling by finding the parens in the first string
				String scaling = match.toString(1);
				int start = scaling.indexOf('(');
				int end = scaling.indexOf(')', start);
				scaling = scaling.substring(start + 1, end);
				data.set(scaling, "income statement scaling");
				income = match.toString(2);
				processAnnualData(income, data);
			}
			else {
				parseErr = "Income statement section not found";
				return false;
			}
		
			// balance sheet section
			String balance;
			match = secParser.getSection("(Balance Sheet.+?)<table.*?>(.+?)</table>");
			if (match != null) {
				// get scaling by finding the parens in the first string
				String scaling = match.toString(1);
				int start = scaling.indexOf('(');
				int end = scaling.indexOf(')', start);
				scaling = scaling.substring(start + 1, end);
				data.set(scaling, "balance sheet scaling");
				balance = match.toString(2);
				processAnnualData(balance, data);
			}
			else {
				parseErr = "Balance sheet section not found";
				return false;
			}

			// ratios section
			String ratio;
			match = secParser.getSection("<center>.*?<table.*?>(.+?%.+?)</table>.*?</center>");
			if (match != null) {
				ratio = match.toString(1);
				processAnnualData(ratio, data);
			}
			else {
				parseErr = "Ratios section not found";
				return false;
			}

			// misc section
			String misc;
			match = secParser.getSection("<font.*?(<B>.+?)This report");
			if (match != null)
				misc = match.toString(1);
			else {
				parseErr = "Miscellaneous section not found";
				return false;
			}
		

			if (DEBUG) {
				//block("HEADER", header);
				//block("INDUSTRY", industry);
				//block("PRICE", price);
				//block("EVALUATION", evaluation);
				//block("OVERVIEW", overview);
				//block("VALUATION", valuation);
				//block("KEY STATS", keyStats);
				//block("REVENUES", revenues);
				//block("EPS", eps);
				//block("PER SHARE DATA", perShare);
				//block("INCOME STATEMENT", income);
				//block("BALANCE SHEET", balance);
				//block("RATIO", ratio);
				//block("MISC", misc);
			
			}
			return true;
		}
		catch (REException e) {
			System.err.println("An error has occurred in the regex engine: " + e.toString());
			parseErr = "Internal regex error: " + e.toString();
			return false;
		}
		catch (IOException e) {
			System.err.println("IO Error: " + e.toString());
			parseErr = "File I/O error";
			return false;
		}
	}

	/**
	 * Return string with error information if a parse failed.
	 */
	public String parseError()
	{
		return parseErr;
	}

	/**
	 * Process a string parsed as an annual data section of a report, such as the
	 * Income Statement, Balance Sheet or Ratios.
	 * Add the data to the CompanyInfo object.
	 */
	void processAnnualData(String s, CompanyInfo data) throws REException
	{
		TableRowEnumeration trEnum = new TableRowEnumeration(s);
		// first process the years
		try	{
			// we may have to read through this more than once to get to the year line
			int totalYears = 0;
			REMatch[] elements = null;
			while (totalYears < 1) {
				String trElem = trEnum.nextElement();
				RE tdRegex = new RE("<td.*?>(.*?)</td>", RE.REG_MULTILINE 
									| RE.REG_ICASE | RE.REG_DOT_NEWLINE);
				elements = tdRegex.getAllMatches(trElem);
				// the first element of the table is blank, not a year
				totalYears = elements.length - 1; 	
			}
			String[] years = new String[totalYears];
			RE yearsRegex = new RE("(\\d{4})");
			for (int i = 1; i < elements.length; i++) {
				REMatch match = yearsRegex.getMatch(elements[i].toString(1));
				years[i - 1] = match.toString(1);
			}

			// now step through until there are no more rows
			while (trEnum.hasMoreElements()) {
				TableDataEnumeration tdEnum = new TableDataEnumeration(trEnum.nextElement());
				// attribute is the key we will use to store a line: earnings, 
				// book value, etc.
				RE exp = new RE("(<.+?>)*\\s*(.+?)\\s*</.+?>");
				REMatch match = exp.getMatch(tdEnum.nextElement());
				String attribute = match.toString(2);
				for (int i = 0; i < totalYears; i++) {
					match = exp.getMatch(tdEnum.nextElement());
					if (match != null)
						data.set(match.toString(2), attribute, years[i]);
					else
						System.err.println("processAnnualData(" + attribute + 
										   ") parse failed for " + years[i]);
				}

			}
		

		}
		catch (java.util.NoSuchElementException e) 	{
			System.err.println("processAnnualData: NoSuchElementException: " + e.getMessage());
			e.printStackTrace();
			return;
		}
	}

	/**
	 * Process a string parsed as the EPS section of a report.
	 * Add the data to the CompanyInfo object.
	 */
	void processEPS(String s, CompanyInfo data) throws REException
	{
		processQuarterlyTable("quarterly eps", s, data);
	}

	/**
	 * Process a string parsed as the evaluation section of a report.
	 * Add the data to the CompanyInfo object.
	 */
	void processEvaluation(String s, CompanyInfo data) throws REException
	{
		// We can't use TableRowEnumeration for this section, because the Waterhouse
		// source is missing the last </tr>
		TableDataEnumeration tde = new TableDataEnumeration(s);
		String elem = tde.nextElement();
		BREnumeration bre = new BREnumeration(elem);
		
		elem = bre.nextElement();
		while (elem.indexOf("Highest") == -1)
			elem = bre.nextElement();
		elem = bre.nextElement();
		RE regex = new RE("<img.+?>\\s*(\\d[+-]?)"
						  , RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
		REMatch match = regex.getMatch(elem);
		if (match != null)
			data.set(match.toString(1), "s&p quantitative evaluation");
		else
			System.err.println("processEvaluation: s&p quantitative evaluation not found");
		
		elem = bre.nextElement();
		elem = bre.nextElement();
		regex = new RE("<img.+?>\\s*((\\d+\\.\\d+))"
					   , RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
		match = regex.getMatch(elem);
		if (match != null)
			data.set(match.toString(1), "s&p fair value");
		else
			System.err.println("processEvaluation: s&p fair value not found");
		
		elem = bre.nextElement();
		elem = bre.nextElement();
		regex = new RE("<img.+?>\\s*(.+?)<", 
					   RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
		match = regex.getMatch(elem);
		if (match != null)
			data.set(match.toString(1), "s&p risk");
		else
			System.err.println("processEvaluation: s&p risk not found");

		elem = bre.nextElement();
		elem = bre.nextElement();
		regex = new RE("<img.+?>\\s*(.[+-]?)", 
					   RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
		match = regex.getMatch(elem);
		if (match != null)
			data.set(match.toString(1), "s&p earn/div rank");
		else
			System.err.println("processEvaluation: s&p earn/div rank not found");

		elem = tde.nextElement();
		while (elem.indexOf("Technical") == -1)
			elem = tde.nextElement();
		
		regex = new RE("<img.+?>\\s*(\\w+?)\\W*?</b>\\s*(since\\s+\\d+/\\d+)",
					   RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
		match = regex.getMatch(elem);
		if (match != null)
			data.set(match.toString(1) + " " + match.toString(2), "s&p technical evaluation");
		else
			System.err.println("processEvaluation: s&p technical evaluation not found");
	
		regex = new RE("Highest.+?<img.+?>\\s*(\\d+)\\D+", 
					   RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
		match = regex.getMatch(elem);
		if (match != null)
			data.set(match.toString(1), "s&p relative strength rank");
		else
			System.err.println("processEvaluation: s&p relative strength rank not found");

		regex = new RE("Insider.+?<img.+?>\\s*(\\w+?)\\W", 
					   RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
		match = regex.getMatch(elem);
		if (match != null)
			data.set(match.toString(1), "s&p insider activity");
		else
			System.err.println("processEvaluation: s&p insider activity not found");
	}

	/**
	 * Process a string parsed as the header section of a report.  Add
	 * the data to the CompanyInfo object.  
	 */
	void processHeader(String s, CompanyInfo data) throws REException
	{
		try	{
			TableDataEnumeration enum = new TableDataEnumeration(s);
			String elem = enum.nextElement();
			while (elem.indexOf("EPORTS") == -1)
				elem = enum.nextElement();

			elem = enum.nextElement();
			RE regex = new RE("<font.*?>(.+?)</font>", 
							  RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
			REMatch match = regex.getMatch(elem);
			// this part is a bit of a repeat, but the <TITLE> section is broken 
			// at Waterhouse
			if (match != null)
				data.set(match.toString(1), "name");
			else
				System.out.println("processHeader: name search failed");
			elem = enum.nextElement();
			regex = new RE("\\s*(\\w+)\\s+Symbol\\s*<font.*?>\\s*(.+?)\\s*</font>", 
			    RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
			match = regex.getMatch(elem);
			if (match != null) {
				data.set(match.toString(1), "exchange");
				data.set(match.toString(2), "ticker");
			}
			else
				System.out.println("processHeader: exchange/ticker search failed");
		}
		catch (java.util.NoSuchElementException e) {
			System.err.println("processHeader: NoSuchElementException");
			return;
		}
	}

	/**
	 * Process a string parsed as the industry section of a report.
	 * Add the data to the CompanyInfo object.
	 */
	void processIndustry(String s, CompanyInfo data) throws REException
	{
		try	{
			TableDataEnumeration enum = new TableDataEnumeration(s);
			// find report date
			RE regex = new RE("(\\d+-\\w+-\\d+)", 
							  RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
			REMatch match;
			String elem = enum.nextElement();
			while ( (match = regex.getMatch(elem)) == null)
				elem = enum.nextElement();
			data.set(match.toString(1), "report date");
		
			// get industry
			regex = new RE("Industry:\\s*(<.+?>)*(\\w.+?)<", 
						   RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
			match = regex.getMatch(enum.nextElement());
			if (match != null)
				data.set(match.toString(2), "industry");
			else
				System.err.println("processIndustry: industry not found");

			// industry summary
			regex = new RE("Summary:\\s*(<.+?>)*(\\w.+?)<", 
						   RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
			match = regex.getMatch(enum.nextElement());
			if (match != null)
				data.set(match.toString(2), "industry summary");
			else
				System.err.println("processIndustry: industry summary not found");
		
		
		}
		catch (java.util.NoSuchElementException e) {
			System.err.println("processIndustry: NoSuchElementException");
			return;
		}
	}

	/**
	 * Process a string parsed as the Key Stock Statistics section of a report.
	 * Add the data to the CompanyInfo object.
	 */
	void processKeyStats(String s, CompanyInfo data) throws REException
	{
		TableDataEnumeration tde = new TableDataEnumeration(s);
		AttributeTranslator attrTrans = new AttributeTranslator(data);
		while (tde.hasMoreElements()) {
			// labels are all bolded
			RE regex = new RE("<b>\\s*(.+?)\\s*</b>", 
							  RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
			REMatch match = regex.getMatch(tde.nextElement());
			if (match != null) {
				String attr = match.toString(1);
				regex = new RE("<font.+?>\\s*(.+?)\\s*</font>", 
							   RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
				match = regex.getMatch(tde.nextElement());
				if (match != null)
					attrTrans.addData(match.toString(1), attr);
			}		
		}
	}

	/**
	 * Process a string parsed as the Per Share Data section of a report.
	 * Add the data to the CompanyInfo object.  
	 */
	void processPerShareData(String s, CompanyInfo data) throws REException {

		try {
			// enumerate through rows of the table
			TableRowEnumeration trEnum = new TableRowEnumeration(s);

			// start with years
			String trElem = trEnum.nextElement();
			RE tdRegex = new RE("<td.*?>(.*?)</td>", 
								RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
			REMatch[] elements = tdRegex.getAllMatches(trElem);
			// the first element of the table is blank, not a year
			int totalYears = elements.length - 1; 	

			String[] years = new String[totalYears];
			RE yearsRegex = new RE("(\\d{4})");
			for (int i = 1; i < elements.length; i++) {
				REMatch match = yearsRegex.getMatch(elements[i].toString(1));
				years[i - 1] = match.toString(1);
			}
			// now loop through the data until we get to either a prices line or
			// a calendar years line
			String attribute = null;
			String currentLine = null;
			boolean readyForPrices = false;
			while (!readyForPrices) {
				// we have to keep this for when we exit the loop
				currentLine = trEnum.nextElement();	
				TableDataEnumeration tdEnum = new TableDataEnumeration(currentLine);
				// attribute is the key we will use to store a line: earnings, 
				// book value, etc.
				RE exp = new RE("(<.+?>)*\\s*(.+?)\\s*</.+?>");
				REMatch match = exp.getMatch(tdEnum.nextElement());
				attribute = match.toString(2);
				if ((attribute.indexOf("Cal") != -1) 
					|| (attribute.indexOf("Price") != -1)) {
					readyForPrices = true;
					continue;
				}
				if ((attribute.indexOf("Cash Flow") != -1))
					attribute = "Cash Flow Per Share";
				for (int i = 0; i < totalYears; i++) {
					match = exp.getMatch(tdEnum.nextElement());
					if (match != null)
						data.set(match.toString(2), attribute, years[i]);
					else
						System.err.println("Per share data (" + attribute + 
										   ") parse failed for " + years[i]);
				}
			}

			// Now process prices.  If the current line contains Cal Yrs, we have 
			// to re-do the years[] array
			if (attribute.indexOf("Cal") != -1)	{
				elements = tdRegex.getAllMatches(currentLine);
				// the first element of the table is not a year
				totalYears = elements.length - 1; 	
				years = new String[totalYears];
				for (int i = 1; i < elements.length; i++) {
					REMatch match = yearsRegex.getMatch(elements[i].toString(1));
					years[i - 1] = match.toString(1);
				}
				currentLine = trEnum.nextElement();
			}
		
			// step through 4 lines of price and p/e data
			// currentLine is already set to a line to be processed
			for (int i = 1; i <= 4; i++) {
				// attribute is the key we will use to store a line: earnings, 
				// book value, etc.
				if (i < 3)
					attribute = "Price - ";
				else
					attribute = "P/E Ratio - ";
				if ((i % 2) == 1)
					attribute += "High";
				else
					attribute += "Low";
			
				TableDataEnumeration tdEnum = new TableDataEnumeration(currentLine);
				// throw away first element, since we're setting attribute based 
				// on the value of i
				tdEnum.nextElement();
				RE exp = new RE("(<.+?>)*\\s*(.+?)\\s*</.+?>");
				for (int j = 0; j < totalYears; j++) {
					REMatch match = exp.getMatch(tdEnum.nextElement());
					if (match != null)
						data.set(match.toString(2), attribute, years[j]);
					else
						System.err.println("Per share data (" + attribute + 
										   ") parse failed for " + years[j]);
				}
				if (i < 4)
					currentLine = trEnum.nextElement();
			}
		
		} catch (java.util.NoSuchElementException e) {
			System.err.println("processPerShareData: NoSuchElementException: " 
							   + e.getMessage());
			e.printStackTrace();
			return;
		}
	}

	/**
	 * Process the price info block
	 * @param data java.lang.String
	 * @param data2 sptossg.CompanyInfo
	 */
	void processPrice(String s, CompanyInfo data) throws REException
	{
		TableDataEnumeration tde = new TableDataEnumeration(s);
		String elem = tde.nextElement();
		elem = tde.nextElement();

		// current price and 52 week high/low
		RE regex = new RE("Recent Price\\s+<img.+?>\\s*(\\d+\\.\\d+).+?<img.+?>\\s*(\\d+\\.\\d+) - (\\d+\\.\\d+)", RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
		REMatch match = regex.getMatch(elem);
		if (match != null) {
			data.set(match.toString(1), "current price");
			data.set(match.toString(2), "52 week high price");
			data.set(match.toString(3), "52 week low price");
		}
		else
			System.err.println("processPrice: current price not found");

		elem = tde.nextElement();
		regex = new RE("Yield\\s*<img.+?>\\s*(.+?)%?<br>.+?<img.+?>\\s*(.+?)</font>", 
					   RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
		match = regex.getMatch(elem);
		if (match != null) {
			String temp = match.toString(1);
			if (temp.equalsIgnoreCase("Nil"))
				data.set("0", "yield");
			else
				data.set(temp, "yield");
			data.set(match.toString(2), "12 month p/e");
		}
		else
			System.err.println("processPrice: yield and 12 month p/e not found");
	}

	/**
	 * Process a table of quarterly data.
	 * @param attribute Attribute name for quarterly data to be stored
	 * @param s String containing the table to be parsed
	 * @param data CompanyInfo object for storing the data  
	 */
	void processQuarterlyTable(String attribute, String s, CompanyInfo data) 
		throws REException
	{
		try	{
			// enumerate through rows of the table
			TableRowEnumeration trEnum = new TableRowEnumeration(s);

			// start with years
			String trElem = trEnum.nextElement();
			RE tdRegex = new RE("<td.*?>(.*?)</td>", 
								RE.REG_MULTILINE | RE.REG_ICASE | RE.REG_DOT_NEWLINE);
			REMatch[] elements = tdRegex.getAllMatches(trElem);
			// the first element of the table is blank, not a year
			int totalYears = elements.length - 1; 
			String[] years = new String[totalYears];
			RE yearsRegex = new RE("(\\d{4})");
			for (int i = 1; i < elements.length; i++) {
				REMatch match = yearsRegex.getMatch(elements[i].toString(1));
				years[i - 1] = match.toString(1);
			}
			// now loop through the quarters
			for (int i = 1; i <= 4; i++) {
				TableDataEnumeration tdEnum = new TableDataEnumeration(trEnum.nextElement());
				tdEnum.nextElement();
				String qtr = "Q" + i;
				for (int j = 0; j < totalYears; j++) {
					RE exp = new RE("(<.+?>)*\\s*(.+?)\\s*</.+?>");
					String st = tdEnum.nextElement();
					REMatch m = exp.getMatch(st);
					if (m != null)
						data.set(m.toString(2), attribute, years[j], qtr);
					else
						System.err.println("Quarterly " + attribute + 
										   " parse failed for " + years[j] + " " + qtr);
				}
			}
		}
		catch (java.util.NoSuchElementException e) {
			System.err.println("processQuarterlyTable: NoSuchElementException");
			return;
		}
	}

	/**
	 * Process a string parsed as the revenues section of a report.
	 * Add the data to the CompanyInfo object.
	 */
	void processRevenues(String s, CompanyInfo data) throws REException
	{
		processQuarterlyTable("quarterly revenues", s, data);
	}

}

